# -*- coding: utf-8 -*-
"""
Test select one external syntax.
"""
import codecs
import os

import unittest2 as unittest

import pyxform
from pyxform.tests import utils
from pyxform.tests.utils import XFormTestCase
from pyxform.utils import sheet_to_csv
from pyxform import InputQuestion, Survey

DIR = os.path.dirname(__file__)


class MainTest(XFormTestCase):

    maxDiff = None

    def runTest(self):
        for filename in ["select_one_external.xlsx"]:
            self.get_file_path(filename)
            expected_output_path = os.path.join(
                DIR, "test_expected_output", self.root_filename + ".xml"
            )

            output_csv = os.path.join(DIR, "test_output", self.root_filename + ".csv")
            # Do the conversion:
            json_survey = pyxform.xls2json.parse_file_to_json(
                self.path_to_excel_file, default_name="select_one_external"
            )

            self.assertTrue(
                sheet_to_csv(self.path_to_excel_file, output_csv, "external_choices")
            )
            self.assertFalse(
                sheet_to_csv(self.path_to_excel_file, output_csv, "non-existant sheet")
            )

            survey = pyxform.create_survey_element_from_dict(json_survey)

            survey.print_xform_to_file(self.output_path)

            # Compare with the expected output:
            with codecs.open(
                expected_output_path, "rb", encoding="utf-8"
            ) as expected_file:
                with codecs.open(
                    self.output_path, "rb", encoding="utf-8"
                ) as actual_file:
                    self.assertXFormEqual(expected_file.read(), actual_file.read())

    def test_csv_data(self):
        """
        Test that the data on the external choices sheet is cleaned and written
        appropriately to the csv file.
        """

        filename = "select_one_external.xlsx"
        self.get_file_path(filename)

        output_csv = os.path.join(DIR, "test_output", self.root_filename + ".csv")

        self.assertTrue(
            sheet_to_csv(self.path_to_excel_file, output_csv, "external_choices")
        )

        # Compare with the expected output
        with codecs.open(output_csv, "rb", encoding="utf-8") as expected_file:
            rows = expected_file.read()
            # Evaluate the last line obtained from the csv file
            self.assertEqual(
                rows.splitlines()[-1],
                '"cities","puyallup","Puyallup","washington","pierce"',
            )
            # Evaluate the first line obtained from the csv file
            self.assertEqual(
                rows.splitlines()[0], '"list_name","name","label","state","county"'
            )

    def test_output_node_for_select_one_question_type(self):

        self.this_directory = os.path.dirname(__file__)
        survey_out = Survey(name="geopgraphy", sms_keyword="geography", type="survey")
        question = InputQuestion(name="counties")
        question.type = "select one external"
        question.label = "county"
        survey_out.add_child(question)

        expected = '<input ref="/geopgraphy/counties">'

        xml = survey_out.to_xml()
        self.assertEqual(1, xml.count(expected))


if __name__ == "__main__":
    unittest.main()



# -*- coding: utf-8 -*-
"""
Testing creation of Surveys using verbose methods
"""
from unittest import TestCase

from pyxform import Survey
from pyxform.builder import create_survey_element_from_dict
from pyxform.tests.utils import prep_class_config

TESTING_BINDINGS = True


def ctw(control):
    """
    ctw stands for control_test_wrap, but ctw is shorter and easier. using
    begin_str and end_str to take out the wrap that xml gives us
    """
    return control.toxml()


class Json2XformQuestionValidationTests(TestCase):
    maxDiff = None

    @classmethod
    def setUpClass(cls):
        prep_class_config(cls=cls)

    def setUp(self):
        self.s = Survey(name="test")

    def test_question_type_string(self):
        simple_string_json = {
            "label": {
                "French": "Nom du travailleur agricole:",
                "English": "Name of Community Agricultural Worker",
            },
            "type": "text",
            "name": "enumerator_name",
        }

        q = create_survey_element_from_dict(simple_string_json)

        expected_string_control_xml = self.config.get(
            self.cls_name, "test_question_type_string_control"
        )

        expected_string_binding_xml = self.config.get(
            self.cls_name, "test_question_type_string_binding"
        )

        self.s.add_child(q)
        self.assertEqual(ctw(q.xml_control()), expected_string_control_xml)

        if TESTING_BINDINGS:
            self.assertEqual(ctw(q.xml_binding()), expected_string_binding_xml)

    def test_select_one_question_multilingual(self):
        """
        Test the lowest common denominator of question types.
        """
        simple_select_one_json = {
            "label": {"f": "ftext", "e": "etext"},
            "type": "select one",
            "name": "qname",
            "choices": [
                {"label": {"f": "fa", "e": "ea"}, "name": "a"},
                {"label": {"f": "fb", "e": "eb"}, "name": "b"},
            ],
        }

        # I copied the response in, since this is not our method of testing
        # valid return values.
        expected_select_one_control_xml = self.config.get(
            self.cls_name, "test_select_one_question_multilingual_control"
        )

        expected_select_one_binding_xml = self.config.get(
            self.cls_name, "test_select_one_question_multilingual_binding"
        )

        q = create_survey_element_from_dict(simple_select_one_json)
        self.s.add_child(q)
        self.assertEqual(ctw(q.xml_control()), expected_select_one_control_xml)

        if TESTING_BINDINGS:
            self.assertEqual(ctw(q.xml_binding()), expected_select_one_binding_xml)

    def test_simple_integer_question_type_multilingual(self):
        """
        not sure how integer questions should show up.
        """
        simple_integer_question = {
            "label": {"f": "fc", "e": "ec"},
            "type": "integer",
            "name": "integer_q",
            "attributes": {},
        }

        expected_integer_control_xml = self.config.get(
            self.cls_name, "test_simple_integer_question_type_multilingual_control"
        )

        expected_integer_binding_xml = self.config.get(
            self.cls_name, "test_simple_integer_question_type_multilingual_binding"
        )

        q = create_survey_element_from_dict(simple_integer_question)

        self.s.add_child(q)

        self.assertEqual(ctw(q.xml_control()), expected_integer_control_xml)

        if TESTING_BINDINGS:
            self.assertEqual(ctw(q.xml_binding()), expected_integer_binding_xml)

    def test_simple_date_question_type_multilingual(self):
        """
        not sure how date questions should show up.
        """
        simple_date_question = {
            "label": {"f": "fd", "e": "ed"},
            "type": "date",
            "name": "date_q",
            "attributes": {},
        }

        expected_date_control_xml = self.config.get(
            self.cls_name, "test_simple_date_question_type_multilingual_control"
        )

        expected_date_binding_xml = self.config.get(
            self.cls_name, "test_simple_date_question_type_multilingual_binding"
        )

        q = create_survey_element_from_dict(simple_date_question)
        self.s.add_child(q)
        self.assertEqual(ctw(q.xml_control()), expected_date_control_xml)

        if TESTING_BINDINGS:
            self.assertEqual(ctw(q.xml_binding()), expected_date_binding_xml)

    def test_simple_phone_number_question_type_multilingual(self):
        """
        not sure how phone number questions should show up.
        """
        simple_phone_number_question = {
            "label": {"f": "fe", "e": "ee"},
            "type": "phone number",
            "name": "phone_number_q",
        }

        expected_phone_number_control_xml = self.config.get(
            self.cls_name, "test_simple_phone_number_question_type_multilingual_control"
        )

        expected_phone_number_binding_xml = self.config.get(
            self.cls_name, "test_simple_phone_number_question_type_multilingual_binding"
        )

        q = create_survey_element_from_dict(simple_phone_number_question)
        self.s.add_child(q)
        self.assertEqual(ctw(q.xml_control()), expected_phone_number_control_xml)

        if TESTING_BINDINGS:
            self.assertEqual(ctw(q.xml_binding()), expected_phone_number_binding_xml)

    def test_simple_select_all_question_multilingual(self):
        """
        not sure how select all questions should show up...
        """
        simple_select_all_question = {
            "label": {"f": "f choisit", "e": "e choose"},
            "type": "select all that apply",
            "name": "select_all_q",
            "choices": [
                {"label": {"f": "ff", "e": "ef"}, "name": "f"},
                {"label": {"f": "fg", "e": "eg"}, "name": "g"},
                {"label": {"f": "fh", "e": "eh"}, "name": "h"},
            ],
        }

        expected_select_all_control_xml = self.config.get(
            self.cls_name, "test_simple_select_all_question_multilingual_control"
        )

        expected_select_all_binding_xml = self.config.get(
            self.cls_name, "test_simple_select_all_question_multilingual_binding"
        )

        q = create_survey_element_from_dict(simple_select_all_question)
        self.s.add_child(q)
        self.assertEqual(ctw(q.xml_control()), expected_select_all_control_xml)

        if TESTING_BINDINGS:
            self.assertEqual(ctw(q.xml_binding()), expected_select_all_binding_xml)

    def test_simple_decimal_question_multilingual(self):
        """
        not sure how decimal should show up.
        """
        simple_decimal_question = {
            "label": {"f": "f text", "e": "e text"},
            "type": "decimal",
            "name": "decimal_q",
            "attributes": {},
        }

        expected_decimal_control_xml = self.config.get(
            self.cls_name, "test_simple_decimal_question_multilingual_control"
        )

        expected_decimal_binding_xml = self.config.get(
            self.cls_name, "test_simple_decimal_question_multilingual_binding"
        )

        q = create_survey_element_from_dict(simple_decimal_question)
        self.s.add_child(q)
        self.assertEqual(ctw(q.xml_control()), expected_decimal_control_xml)

        if TESTING_BINDINGS:
            self.assertEqual(ctw(q.xml_binding()), expected_decimal_binding_xml)



# -*- coding: utf-8 -*-
"""
Test handling dynamic default in forms
"""

from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class DynamicDefaultTests(PyxformTestCase):
    """
    Handling dynamic defaults
    """

    def test_handling_dynamic_default(self):
        """
        Should use set-value for dynamic default form
        """
        md = """
        | survey |         |            |            |                   |
        |        | type    | name       | label      | default           |
        |        | text    | last_name  | Last name  | not_func$         |
        |        | integer | age        | Your age   | random() |
        """

        self.assertPyxformXform(
            md=md,
            name="dynamic",
            id_string="id",
            model__contains=[
                "<last_name>not_func$</last_name>",
                "<age/>",
                '<setvalue event="odk-instance-first-load" ref="/dynamic/age" value="random()"/>',
            ],
        )

        survey = self.md_to_pyxform_survey(
            md_raw=md,
            kwargs={"id_string": "id", "name": "dynamic", "title": "some-title"},
            autoname=False,
        )
        survey_xml = survey._to_pretty_xml()

        self.assertContains(survey_xml, "<last_name>not_func$</last_name>", 1)
        self.assertContains(survey_xml, "<age/>", 1)
        self.assertContains(
            survey_xml,
            '<setvalue event="odk-instance-first-load" ref="/dynamic/age" value="random()"/>',
            1,
        )
        self.assertNotContains(
            survey_xml,
            '<setvalue event="odk-instance-first-load odk-new-repeat" ref="/dynamic/age" value="random()"/>',
        )

    def test_static_defaults(self):
        self.assertPyxformXform(
            name="dynamic",
            md="""
            | survey |              |          |       |                   |
            |        | type         | name     | label | default           |
            |        | integer      | foo      | Foo   | foo               |
            |        | begin repeat | repeat   |       |                   |
            |        | integer      | bar      | Bar   | 12                |
            |        | end repeat   | repeat   |       |                   |
            """,
            model__contains=["<foo>foo</foo>", "<bar>12</bar>"],
            model__excludes=["setvalue"],
        )

    def test_handling_dynamic_default_in_repeat(self):
        """
        Should use set-value for dynamic default form inside repeat
        """
        md = """
        | survey |              |            |              |                   |
        |        | type         | name       | label        | default           |
        |        | begin repeat | household  | Households   |                   |
        |        | integer      | age        | Your age     | random() |
        |        | text         | feeling    | Your feeling | not_func$         |
        |        | end repeat   | household  |              |                   |
        """

        self.assertPyxformXform(
            md=md,
            name="dynamic",
            id_string="id",
            model__contains=["<age/>", "<feeling>not_func$</feeling>"],
            model__excludes=[
                '<setvalue event="odk-instance-first-load odk-new-repeat" ref="/dynamic/household/age" value="random()"/>'
            ],
            xml__contains=[
                '<setvalue event="odk-instance-first-load odk-new-repeat" ref="/dynamic/household/age" value="random()"/>'
            ],
        )

        survey = self.md_to_pyxform_survey(
            md_raw=md,
            kwargs={"id_string": "id", "name": "dynamic", "title": "some-title"},
            autoname=False,
        )
        survey_xml = survey._to_pretty_xml()

        self.assertContains(survey_xml, "<feeling>not_func$</feeling>", 2)
        self.assertContains(survey_xml, "<age/>", 2)
        self.assertContains(
            survey_xml,
            '<setvalue event="odk-instance-first-load odk-new-repeat" ref="/dynamic/household/age" value="random()"/>',
            1,
        )
        self.assertNotContains(
            survey_xml,
            '<setvalue event="odk-instance-first-load" ref="/dynamic/age" value="random()"/>',
        )

    def test_dynamic_default_in_group(self):
        self.assertPyxformXform(
            name="dynamic",
            md="""
            | survey |            |          |       |                   |
            |        | type       | name     | label | default           |
            |        | integer    | foo      | Foo   |                   |
            |        | begin group| group    |       |                   |
            |        | integer    | bar      | Bar   | ${foo}            |
            |        | end group  | group    |       |                   |
            """,
            model__contains=[
                '<setvalue event="odk-instance-first-load" ref="/dynamic/group/bar" value=" /dynamic/foo "/>'
            ],
        )

    def test_sibling_dynamic_default_in_group(self):
        self.assertPyxformXform(
            name="dynamic",
            md="""
            | survey |              |          |       |                   |
            |        | type         | name     | label | default           |
            |        | begin group  | group    |       |                   |
            |        | integer      | foo      | Foo   |                   |
            |        | integer      | bar      | Bar   | ${foo}            |
            |        | end group    | group    |       |                   |
            """,
            model__contains=[
                '<setvalue event="odk-instance-first-load" ref="/dynamic/group/bar" value=" /dynamic/group/foo "/>'
            ],
        )

    def test_sibling_dynamic_default_in_repeat(self):
        self.assertPyxformXform(
            name="dynamic",
            md="""
            | survey |              |          |       |                   |
            |        | type         | name     | label | default           |
            |        | begin repeat | repeat   |       |                   |
            |        | integer      | foo      | Foo   |                   |
            |        | integer      | bar      | Bar   | ${foo}            |
            |        | end repeat   | repeat   |       |                   |
            """,
            xml__contains=[
                '<setvalue event="odk-instance-first-load odk-new-repeat" ref="/dynamic/repeat/bar" value=" ../foo "/>'
            ],
            model__excludes=[
                '<setvalue event="odk-instance-first-load odk-new-repeat" ref="/dynamic/repeat/bar" value=" ../foo "/>'
            ],
        )

    def test_dynamic_default_in_group_nested_in_repeat(self):
        self.assertPyxformXform(
            name="dynamic",
            md="""
            | survey |              |          |       |                   |
            |        | type         | name     | label | default           |
            |        | begin repeat | repeat   |       |                   |
            |        | begin group  | group    |       |                   |
            |        | integer      | foo      | Foo   |                   |
            |        | integer      | bar      | Bar   | ${foo}            |
            |        | end group    | group    |       |                   |
            |        | end repeat   | repeat   |       |                   |
            """,
            xml__contains=[
                '<setvalue event="odk-instance-first-load odk-new-repeat" ref="/dynamic/repeat/group/bar" value=" ../foo "/>'
            ],
            model__excludes=['<setvalue event="odk-instance-first-load'],
        )

    def test_dynamic_defaults_in_nested_repeat(self):
        md = """
            | survey |              |          |       |                   |
            |        | type         | name     | label | default           |
            |        | begin repeat | outer    |       |                   |
            |        | date         | date     | Date  | now()             |
            |        | integer      | foo      | Foo   |                   |
            |        | begin repeat | inner    |       |                   |
            |        | integer      | bar      | Bar   | ${foo}            |
            |        | end repeat   | inner    |       |                   |
            |        | end repeat   | outer    |       |                   |
            """

        self.assertPyxformXform(
            name="dynamic",
            md=md,
            xml__contains=[
                '<setvalue event="odk-instance-first-load odk-new-repeat" ref="/dynamic/outer/inner/bar" value=" ../../foo "/>',
                '<setvalue event="odk-instance-first-load odk-new-repeat" ref="/dynamic/outer/date" value="now()"/>',
            ],
            model__excludes=['<setvalue event="odk-instance-first-load'],
        )

        survey = self.md_to_pyxform_survey(
            md_raw=md,
            kwargs={"id_string": "id", "name": "dynamic", "title": "some-title"},
            autoname=False,
        )
        survey_xml = survey._to_pretty_xml()

        self.assertContains(
            survey_xml,
            '<setvalue event="odk-instance-first-load odk-new-repeat" ref="/dynamic/outer/inner/bar" value=" ../../foo "/>',
            1,
        )

    def test_handling_arithmetic_expression(self):
        """
        Should use set-value for dynamic default form
        """
        md = """
        | survey |         |            |             |                   |
        |        | type    | name       | label       | default           |
        |        | text    | expr_1     | First expr  | 2 + 3 * 4         |
        |        | text    | expr_2     | Second expr | 5 div 5 - 5       |
        |        | integer | expr_3     | Third expr  | random() + 2 * 5    |
        """

        self.assertPyxformXform(
            md=md,
            name="dynamic",
            id_string="id",
            model__contains=[
                "<expr_1/>",
                "<expr_2/>",
                "<expr_3/>",
                '<setvalue event="odk-instance-first-load" ref="/dynamic/expr_1" value="2 + 3 * 4"/>',
                '<setvalue event="odk-instance-first-load" ref="/dynamic/expr_2" value="5 div 5 - 5"/>',
                '<setvalue event="odk-instance-first-load" ref="/dynamic/expr_3" value="random() + 2 * 5"/>',
            ],
        )

    def test_handling_arithmetic_text(self):
        """
        Should use set-value for dynamic default form with arithmetic text
        """
        md = """
        | survey |         |            |             |         |
        |        | type    | name       | label       | default |
        |        | text    | expr_1     | First expr  | 3 mod 3 |
        |        | text    | expr_2     | Second expr | 5 div 5 |
        """

        self.assertPyxformXform(
            md=md,
            name="dynamic",
            id_string="id",
            model__contains=[
                "<expr_1/>",
                "<expr_2/>",
                '<setvalue event="odk-instance-first-load" ref="/dynamic/expr_1" value="3 mod 3"/>',
                '<setvalue event="odk-instance-first-load" ref="/dynamic/expr_2" value="5 div 5"/>',
            ],
        )

    def test_dynamic_default_with_nested_expression(self):
        self.assertPyxformXform(
            name="dynamic",
            md="""
            | survey |         |               |               |                   |
            |        | type    | name          | label         | default           |
            |        | integer | patient_count | Patient count | if(${last-saved#patient_count} = '', 0, ${last-saved#patient_count} + 1) |
            """,
            xml__contains=[
                '<instance id="__last-saved" src="jr://instance/last-saved"/>',
                '<setvalue event="odk-instance-first-load" ref="/dynamic/patient_count" '
                "value=\"if( instance('__last-saved')/dynamic/patient_count  = '', 0,  "
                "instance('__last-saved')/dynamic/patient_count  + 1)\"/>",
            ],
        )

    def test_dynamic_default_with_reference(self):
        self.assertPyxformXform(
            name="dynamic",
            md="""
            | survey |            |          |       |                   |
            |        | type       | name     | label | default           |
            |        | integer    | foo      | Foo   |                   |
            |        | integer    | bar      | Bar   | ${foo}            |
            """,
            xml__contains=[
                '<setvalue event="odk-instance-first-load" ref="/dynamic/bar" value=" /dynamic/foo "/>'
            ],
        )

    def test_dynamic_default_warns(self):
        warnings = []

        self.md_to_pyxform_survey(
            """
            | survey |      |         |       |         |
            |        | type | name    | label | default |
            |        | text | foo     | Foo   |         |
            |        | text | bar     | Bar   | ${foo}  |
            """,
            warnings=warnings,
        )

        self.assertTrue(len(warnings) == 0)

    def test_default_date_not_considered_dynamic(self):
        self.assertPyxformXform(
            name="dynamic",
            md="""
            | survey |            |          |       |                   |
            |        | type       | name     | label | default           |
            |        | date       | foo      | Foo   | 2020-01-01        |
            """,
            xml__contains=["<foo>2020-01-01</foo>"],
        )

    def test_dynamic_default_on_calculate(self):
        self.assertPyxformXform(
            name="dynamic",
            md="""
            | survey |            |          |       |             |                      |
            |        | type       | name     | label | calculation | default              |
            |        | calculate  | r        |       |             | random() + 0.5       |
            |        | calculate  | one      |       |             | if(${r} < 1,'A','B') |
            """,
            xml__contains=[
                """<setvalue event="odk-instance-first-load" ref="/dynamic/r" value="random() + 0.5"/>""",
                """<setvalue event="odk-instance-first-load" ref="/dynamic/one" value="if( /dynamic/r  &lt; 1,'A','B')"/>""",
            ],
        )



# -*- coding: utf-8 -*-
"""
Test image max-pixel parameters.
"""
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class MaxPixelsTest(PyxformTestCase):
    """
    Test image max-pixel parameters.
    """

    def test_integer_max_pixels(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |        |          |       |                |
            |        | type   | name     | label | parameters     |
            |        | image  | my_image | Image | max-pixels=640 |
            """,
            xml__contains=[
                'xmlns:orx="http://openrosa.org/xforms"',
                '<bind nodeset="/data/my_image" type="binary" orx:max-pixels="640"/>',
            ],
        )

    def test_string_max_pixels(self):
        self.assertPyxformXform(
            name="data",
            errored=True,
            md="""
            | survey |        |          |       |                |
            |        | type   | name     | label | parameters     |
            |        | image  | my_image | Image | max-pixels=foo |
            """,
            error__contains=["Parameter max-pixels must have an integer value."],
        )

    def test_string_extra_params(self):
        self.assertPyxformXform(
            name="data",
            errored=True,
            md="""
            | survey |        |          |       |                        |
            |        | type   | name     | label | parameters             |
            |        | image  | my_image | Image | max-pixels=640 foo=bar |
            """,
            error__contains=[
                "Accepted parameters are 'max-pixels': 'foo' is an invalid parameter."
            ],
        )

    def test_image_with_no_max_pixels_should_warn(self):
        warnings = []

        self.md_to_pyxform_survey(
            """
            | survey |       |            |         |
            |        | type  | name       | label   |
            |        | image | my_image   | Image   |
            |        | image | my_image_1 | Image 1 |
            """,
            warnings=warnings,
        )

        self.assertTrue(len(warnings) == 2)
        self.assertTrue("max-pixels" in warnings[0] and "max-pixels" in warnings[1])



# -*- coding: utf-8 -*-
"""
Test tutorial XLSForm.
"""
from unittest import TestCase

from pyxform.builder import create_survey_from_path
from pyxform.tests import utils


class TutorialTests(TestCase):
    def test_create_from_path(self):
        path = utils.path_to_text_fixture("tutorial.xls")
        create_survey_from_path(path)



# -*- coding: utf-8 -*-
"""
Test that any row with a calculation becomes a calculate of the row's type or of type string if the type is "calculate". A hint or label
error should only be thrown for a row without a calculation.
"""

from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class TypedCalculatesTest(PyxformTestCase):
    def test_xls_type_calculate_has_type_string(self):
        self.assertPyxformXform(
            name="calculate-type",
            md="""
            | survey |          |      |             |             |
            |        | type     | name | label       | calculation |
            |        | calculate| a    |             | 2 * 2       |
            """,
            xml__contains=[
                '<bind calculate="2 * 2" nodeset="/calculate-type/a" type="string"/>',
            ],
        )

    def test_xls_type_calculate_with_label_has_no_body(self):
        self.assertPyxformXform(
            name="calculate-type",
            md="""
                | survey |          |      |             |             |
                |        | type     | name | label       | calculation |
                |        | calculate| a    | A           | 2 * 2       |
                """,
            xml__contains=[
                '<bind calculate="2 * 2" nodeset="/calculate-type/a" type="string"/>',
                "<h:body/>",
            ],
        )

    def test_non_calculate_type_with_calculation_is_bind_type(self):
        self.assertPyxformXform(
            name="non-calculate-type",
            md="""
            | survey |          |      |             |             |
            |        | type     | name | label       | calculation |
            |        | integer  | a    |             | 2 * 2       |
            """,
            xml__contains=[
                '<bind calculate="2 * 2" nodeset="/non-calculate-type/a" type="int"/>'
            ],
        )

    def test_non_calculate_type_with_calculation_and_no_label_has_no_control(self):
        self.assertPyxformXform(
            name="no-label",
            md="""
                | survey |          |      |             |             |
                |        | type     | name | label       | calculation |
                |        | integer  | a    |             | 2 * 2       |
                """,
            instance__contains=["<a/>"],
            xml__excludes=["input"],
        )

    def test_non_calculate_type_with_calculation_no_warns(self):
        warnings = []

        self.md_to_pyxform_survey(
            """
            | survey |           |      |             |      |             |
            |        | type      | name | label       | hint | calculation |
            |        | dateTime  | a    |             |      | now()       |
            |        | integer   | b    |             |      | 1 div 1     |
            |        | note      | note | Hello World |      |             |
            """,
            warnings=warnings,
        )

        self.assertTrue(len(warnings) == 0)

    def test_non_calculate_type_with_hint_and_no_calculation__no_warning(self):
        warnings = []
        self.md_to_pyxform_survey(
            """
            | survey |           |      |             |           |             |
            |        | type      | name | label       | hint      | calculation |
            |        | dateTime  | a    |             |           | now()       |
            |        | integer   | b    |             | Some hint |             |
            |        | note      | note | Hello World |           |             |
            """,
            warnings=warnings,
        )
        self.assertTrue(len(warnings) == 0)

    def test_non_calculate_type_with_calculation_and_dynamic_default_warns(self):
        warnings = []

        self.md_to_pyxform_survey(
            """
            | survey |           |      |             |      |             |         |
            |        | type      | name | label       | hint | calculation | default |
            |        | dateTime  | a    |             |      | now()       |         |
            |        | integer   | b    |             |      | 1 div 1     | $(a)    |
            |        | note      | note | Hello World |      |             |         |
            """,
            warnings=warnings,
        )

        self.assertTrue(len(warnings) == 0)

    def test_non_calculate_type_with_calculation_and_default_no_warns(self):
        warnings = []

        self.md_to_pyxform_survey(
            """
            | survey |           |      |             |      |             |         |
            |        | type      | name | label       | hint | calculation | default |
            |        | dateTime  | a    |             |      | now()       |         |
            |        | integer   | b    |             |      | 1 div 1     | 1       |
            |        | note      | note | Hello World |      |             |         |
            """,
            warnings=warnings,
        )

        self.assertTrue(len(warnings) == 0)

    def test_select_type_with_calculation_and_no_label_has_no_control(self):
        self.assertPyxformXform(
            name="calculate-select",
            md="""
                | survey  |                   |      |       |                  |
                |         | type              | name | label | calculation      |
                |         | select_one yes_no | a    |       | concat('a', 'b') |
                | choices |                   |      |       |                  |
                |         | list_name         | name | label |                  |
                |         | yes_no            | yes  | Yes   |                  |
                |         | yes_no            | no   | No    |                  |
                """,
            xml__contains=[
                '<bind calculate="concat(\'a\', \'b\')" nodeset="/calculate-select/a" type="string"/>'
            ],
            instance__contains=["<a/>"],
            xml__excludes=["<select1>"],
        )

    def test_row_without_label_or_calculation_throws_error(self):
        self.assertPyxformXform(
            name="no-label",
            md="""
        | survey |          |      |             |
        |        | type     | name | label       |
        |        | integer  | a    |             |
        """,
            errored=True,
            error__contains="The survey element named 'a' has no label or hint.",
        )

    def test_calculate_without_calculation_without_default(self):
        self.assertPyxformXform(
            name="calculate-without-calculation-without-default",
            md="""
        | survey |            |      |             |             |         |
        |        | type       | name | label       | calculation | default |
        |        | calculate  | a    |             |             |         |
        """,
            errored=True,
            error__contains="Missing calculation",
        )

    def test_calculate_without_calculation_with_default_without_dynamic_default(self):
        self.assertPyxformXform(
            name="calculate-without-calculation-with-default-without-dynamic-default",
            md="""
        | survey |            |      |             |             |         |
        |        | type       | name | label       | calculation | default |
        |        | calculate  | a    |             |             | foo     |
        """,
            errored=True,
            error__contains="Missing calculation",
        )

    def test_calculate_without_calculation_with_dynamic_default(self):
        self.assertPyxformXform(
            name="calculate-without-calculation-with-dynamic-default",
            md="""
        | survey |            |      |             |             |          |
        |        | type       | name | label       | calculation | default  |
        |        | calculate  | a    |             |             | random() |
        """,
            errored=False,
            instance__contains=["<a/>"],
        )



# -*- coding: utf-8 -*-
"""
Test XForm groups.
"""
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class GroupsTests(PyxformTestCase):
    """
    Test XForm groups.
    """

    def test_group_type(self):
        self.assertPyxformXform(
            md="""
            | survey |             |         |                  |
            |        | type        | name    | label            |
            |        | text        | pregrp  | Pregroup text    |
            |        | begin group | xgrp    | XGroup questions |
            |        | text        | xgrp_q1 | XGroup Q1        |
            |        | integer     | xgrp_q2 | XGroup Q2        |
            |        | end group   |         |                  |
            |        | note        | postgrp | Post group note  |
            """,
            model__contains=[
                "<pregrp/>",
                "<xgrp>",
                "<xgrp_q1/>",  # nopep8
                "<xgrp_q1/>",  # nopep8
                "<xgrp_q2/>",  # nopep8
                "</xgrp>",
                "<postgrp/>",
            ],
        )

    def test_group_intent(self):
        self.assertPyxformXform(
            name="intent_test",
            md="""
            | survey |             |         |                  |                                                             |
            |        | type        | name    | label            | intent                                                      |
            |        | text        | pregrp  | Pregroup text    |                                                             |
            |        | begin group | xgrp    | XGroup questions | ex:org.redcross.openmapkit.action.QUERY(osm_file=${pregrp}) |
            |        | text        | xgrp_q1 | XGroup Q1        |                                                             |
            |        | integer     | xgrp_q2 | XGroup Q2        |                                                             |
            |        | end group   |         |                  |                                                             |
            |        | note        | postgrp | Post group note  |                                                             |
            """,  # nopep8
            xml__contains=[
                '<group intent="ex:org.redcross.openmapkit.action.QUERY(osm_file= /intent_test/pregrp )" ref="/intent_test/xgrp">'  # nopep8
            ],
        )



# -*- coding: utf-8 -*-
"""
ExternalInstance class module
"""
from pyxform.survey_element import SurveyElement


class ExternalInstance(SurveyElement):
    def xml_control(self):
        """
        No-op since there is no associated form control to place under <body/>.

        Exists here because there's a soft abstractmethod in SurveyElement.
        """
        pass



# -*- coding: utf-8 -*-
"""
Testing creation of Surveys using verbose methods
"""
from unittest import TestCase

from pyxform import (
    InputQuestion,
    MultipleChoiceQuestion,
    Survey,
    create_survey_from_xls,
)
from pyxform.tests import utils


class Json2XformVerboseSurveyCreationTests(TestCase):
    def test_survey_can_be_created_in_a_verbose_manner(self):
        s = Survey()
        s.name = "simple_survey"

        q = MultipleChoiceQuestion()
        q.name = "cow_color"
        q.type = "select one"

        q.add_choice(label="Green", name="green")
        s.add_child(q)

        expected_dict = {
            "name": "simple_survey",
            "children": [
                {
                    "name": "cow_color",
                    "type": "select one",
                    "children": [{"label": "Green", "name": "green"}],
                }
            ],
        }
        self.maxDiff = None
        self.assertEqual(s.to_json_dict(), expected_dict)

    def test_survey_can_be_created_in_a_slightly_less_verbose_manner(self):
        option_dict_array = [
            {"name": "red", "label": "Red"},
            {"name": "blue", "label": "Blue"},
        ]

        q = MultipleChoiceQuestion(name="Favorite_Color", choices=option_dict_array)
        q.type = "select one"
        s = Survey(name="Roses_are_Red", children=[q])

        expected_dict = {
            "name": "Roses_are_Red",
            "children": [
                {
                    "name": "Favorite_Color",
                    "type": "select one",
                    "children": [
                        {"label": "Red", "name": "red"},
                        {"label": "Blue", "name": "blue"},
                    ],
                }
            ],
        }

        self.assertEqual(s.to_json_dict(), expected_dict)

    def test_two_options_cannot_have_the_same_value(self):
        q = MultipleChoiceQuestion(name="Favorite Color")
        q.add_choice(name="grey", label="Gray")
        q.add_choice(name="grey", label="Grey")
        self.assertRaises(Exception, q, "validate")

    def test_one_section_cannot_have_two_conflicting_slugs(self):
        q1 = InputQuestion(name="YourName")
        q2 = InputQuestion(name="YourName")
        s = Survey(name="Roses are Red", children=[q1, q2])
        self.assertRaises(Exception, s, "validate")

    def allow_surveys_with_comment_rows(self):
        """assume that a survey with rows that don't have name, type, or label
        headings raise warning only"""
        path = utils.path_to_text_fixture("allow_comment_rows_test.xls")
        survey = create_survey_from_xls(path)
        expected_dict = {
            "default_language": "default",
            "id_string": "allow_comment_rows_test",
            "children": [
                {
                    "name": "farmer_name",
                    "label": {"English": "First and last name of farmer"},
                    "type": "text",
                }
            ],
            "name": "allow_comment_rows_test",
            "_translations": {
                "English": {
                    "/allow_comment_rows_test/farmer_name:label": {
                        "long": "First and last name of farmer"
                    }
                }
            },
            "title": "allow_comment_rows_test",
            "_xpath": {
                "allow_comment_rows_test": "/allow_comment_rows_test",
                "farmer_name": "/allow_comment_rows_test/farmer_name",
            },
            "type": "survey",
        }
        self.assertEquals(survey.to_json_dict(), expected_dict)



# -*- coding: utf-8 -*-
"""
Common base classes for pyxform exceptions.
"""


class PyXFormError(Exception):
    """Common base class for pyxform exceptions."""

    pass


class ValidationError(PyXFormError):
    """Common base class for pyxform validation exceptions."""

    pass



# -*- coding: utf-8 -*-
"""
Test unicode characters in validate error messages.
"""

from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class ValidateUnicodeException(PyxformTestCase):
    """
    Validation errors may include non-ASCII characters. In particular, ODK Validate
    uses ͎ (small arrow) to indicate where a problem starts.
    """

    def test_validate_unicode_exception(self):
        self.assertPyxformXform(
            md="""
            | survey  |           |       |       |                |
            |         | type      | name  | label | calculation    |
            |         | calculate | bad   | bad   | $(myField)='1' |
            """,
            run_odk_validate=True,
            odk_validate_error__contains=[
                'Invalid calculate for the bind attached to "${bad}" : Couldn\'t '
                "understand the expression starting at this point:"
            ],
        )

    def test_validate_with_more_unicode(self):
        self.assertPyxformXform(
            md=u"""
            | survey  |           |       |       |                |
            |         | type      | name  | label | calculation    |
            |         | calculate | bad   | bad   | £¥§©®₱₩        |
            """,
            run_odk_validate=True,
            odk_validate_error__contains=[
                'Invalid calculate for the bind attached to "${bad}" : Couldn\'t '
                "understand the expression starting at this point:"
            ],
        )



# -*- coding: utf-8 -*-
"""
Test settings sheet syntax.
"""

from unittest import TestCase

from pyxform.builder import create_survey_from_path
from pyxform.tests import utils
from pyxform.xls2json import SurveyReader


class SettingsTests(TestCase):

    maxDiff = None

    def setUp(self):
        self.path = utils.path_to_text_fixture("settings.xls")

    def test_survey_reader(self):
        survey_reader = SurveyReader(self.path, default_name="settings")
        expected_dict = {
            u"id_string": u"new_id",
            u"sms_keyword": u"new_id",
            u"default_language": u"default",
            u"name": u"settings",
            u"title": u"My Survey",
            u"type": u"survey",
            u"attribute": {
                u"my_number": u"1234567890",
                u"my_string": u"lor\xe9m ipsum",
            },
            u"children": [
                {
                    u"name": u"your_name",
                    u"label": {u"english": u"What is your name?"},
                    u"type": u"text",
                },
                {
                    u"name": u"your_age",
                    u"label": {u"english": u"How many years old are you?"},
                    u"type": u"integer",
                },
                {
                    "children": [
                        {
                            "bind": {"jr:preload": "uid", "readonly": "true()"},
                            "name": "instanceID",
                            "type": "calculate",
                        }
                    ],
                    "control": {"bodyless": True},
                    "name": "meta",
                    "type": "group",
                },
            ],
        }
        self.assertEqual(survey_reader.to_json_dict(), expected_dict)

    def test_settings(self):
        survey = create_survey_from_path(self.path)
        self.assertEqual(survey.id_string, "new_id")
        self.assertEqual(survey.title, "My Survey")



# -*- coding: utf-8 -*-
"""
InstanceInfo class module.
"""


class InstanceInfo(object):
    """Standardise Instance details relevant during XML generation."""

    def __init__(self, type, context, name, src, instance):
        self.type = type
        self.context = context
        self.name = name
        self.src = src
        self.instance = instance



# -*- coding: utf-8 -*-
"""
The tests utils module functionality.
"""
import os
import xml.etree.ElementTree as ETree

from formencode.doctest_xml_compare import xml_compare
from unittest2 import TestCase
import textwrap
from typing import TYPE_CHECKING

from pyxform import file_utils
from pyxform.builder import create_survey, create_survey_from_path

try:
    import ConfigParser

    configparser = ConfigParser
except ImportError:
    import configparser


if TYPE_CHECKING:
    from typing import Tuple


DIR = os.path.dirname(__file__)


def path_to_text_fixture(filename):
    directory = os.path.dirname(__file__)
    return os.path.join(directory, "example_xls", filename)


def build_survey(filename):
    path = path_to_text_fixture(filename)
    return create_survey_from_path(path)


def create_survey_from_fixture(fixture_name, filetype="xls", include_directory=False):
    fixture_path = path_to_text_fixture("%s.%s" % (fixture_name, filetype))
    noop, section_dict = file_utils.load_file_to_dict(fixture_path)
    pkg = {"main_section": section_dict}
    if include_directory:
        directory, noop = os.path.split(fixture_path)
        pkg["sections"] = file_utils.collect_compatible_files_in_directory(directory)
    return create_survey(**pkg)


class XFormTestCase(TestCase):
    """
    XFormTestCase provides an assertion for comparing two XForms for
    equivalence.  It relies on the xml_compare function provided by FormEncode,
    but with some extra logic to handle the essentially random order of tags in
    the <model> section.  The iteration order of dict keys in Python 3 is
    unpredictable, and many of the elements in <model> tag are generated from
    dicts.  A simple workaround is to just reorder those tags here.  An ideal
    solution might be to rewrite the code that generates an XForm to use
    OrderedDicts where appropriate.
    """

    def get_file_path(self, filename):
        self.path_to_excel_file = os.path.join(DIR, "example_xls", filename)

        # Get the xform output path:
        self.root_filename, self.ext = os.path.splitext(filename)
        self.output_path = os.path.join(DIR, "test_output", self.root_filename + ".xml")

    def assertXFormEqual(self, xform1, xform2):
        xform1 = ETree.fromstring(xform1.encode("utf-8"))
        xform2 = ETree.fromstring(xform2.encode("utf-8"))

        # Sort tags under <model> section in each form
        self.sort_model(xform1)
        self.sort_model(xform2)

        # Report any errors returned from xform_compare
        errs = []

        def reporter(msg):
            errs.append(msg)

        self.assertTrue(
            xml_compare(xform1, xform2, reporter), "\n\n" + "\n".join(reversed(errs))
        )

    def sort_elems(self, elems, attr=None):
        if attr:

            def elem_get_attr(elem):
                return elem.get(attr, "")

            key = elem_get_attr
        else:

            def elem_get_tag(elem):
                return elem.tag

            key = elem_get_tag
        elems[:] = sorted(elems, key=key)

    def sort_model(self, xform):
        ns = "{http://www.w3.org/2002/xforms}"
        model = xform.find(".//" + ns + "model")

        # Sort multiple <instance> tags, if any
        self.sort_elems(model, "id")

        # Sort any <item> tags in each <instance>
        for instance in model.findall(ns + "instance"):
            if instance.get("id", None):
                root = instance.find(ns + "root")
                if root is not None:
                    for item in root:
                        self.sort_elems(item)

        # Sort <translation> tags as well as inner <text> and <value> tags.
        itext = model.find(ns + "itext")
        if itext is None:
            return

        self.sort_elems(itext, "lang")
        for translation in itext:
            self.sort_elems(translation, "id")
            for text in translation:
                self.sort_elems(text, "form")


def prep_class_config(cls, test_dir="tests"):
    cls.config = configparser.ConfigParser()
    here = os.path.dirname(__file__)
    root = os.path.dirname(here)
    strings = os.path.join(root, test_dir, "fixtures", "strings.ini")
    cls.config.read(strings)
    cls.cls_name = cls.__name__


def prep_for_xml_contains(text: str) -> "Tuple[str]":
    """Prep string for finding an exact match to formatted XML text."""
    # noinspection PyRedundantParentheses
    return (textwrap.indent(textwrap.dedent(text), "    ",),)



# -*- coding: utf-8 -*-
"""
Cleans up error messages from the validators.
"""
import re

from pyxform.utils import unicode


class ErrorCleaner(object):
    """Cleans up raw error messages from XForm validators for end users."""

    @staticmethod
    def _replace_xpath_with_tokens(match):
        strmatch = match.group()
        # eliminate e.g /html/body/select1[@ref=/id_string/elId]/item/value
        # instance('q4')/root/item[...]
        if (
            strmatch.startswith("/html/body")
            or strmatch.startswith("/root/item")
            or strmatch.startswith("/html/head/model/bind")
            or strmatch.endswith("/item/value")
        ):
            return strmatch
        line = match.group().split("/")
        return "${%s}" % line[len(line) - 1]

    @staticmethod
    def _cleanup_errors(error_message):
        pattern = r"(/[a-z0-9\-_]+(?:/[a-z0-9\-_]+)+)"
        error_message = re.sub(
            pattern, ErrorCleaner._replace_xpath_with_tokens, error_message, flags=re.I
        )
        lines = unicode(error_message).strip().splitlines()
        no_dupes = [
            line for i, line in enumerate(lines) if line != lines[i - 1] or i == 0
        ]
        return no_dupes

    @staticmethod
    def _remove_java_content(line):
        # has a java filename (with line number)
        has_java_filename = line.find(".java:") != -1
        # starts with '    at java class path or method path'
        is_a_java_method = line.find("\tat") != -1
        if not has_java_filename and not is_a_java_method:
            # remove java.lang.RuntimeException
            if line.startswith("java.lang.RuntimeException: "):
                line = line.replace("java.lang.RuntimeException: ", "")
            # remove org.javarosa.xpath.XPathUnhandledException
            if line.startswith("org.javarosa.xpath.XPathUnhandledException: "):
                line = line.replace("org.javarosa.xpath.XPathUnhandledException: ", "")
            # remove java.lang.NullPointerException
            if line.startswith("java.lang.NullPointerException"):
                line = line.replace("java.lang.NullPointerException", "")
            if line.startswith("org.javarosa.xform.parse.XFormParseException"):
                line = line.replace("org.javarosa.xform.parse.XFormParseException", "")
            return line

    @staticmethod
    def _join_final(error_messages):
        return "\n".join(line for line in error_messages if line is not None)

    @staticmethod
    def odk_validate(error_message):
        if "Error: Unable to access jarfile" in error_message:
            return error_message  # Avoids tokenising the file path.
        common = ErrorCleaner._cleanup_errors(error_message)
        java_clean = [ErrorCleaner._remove_java_content(i) for i in common]
        final_message = ErrorCleaner._join_final(java_clean)
        return final_message

    @staticmethod
    def enketo_validate(error_message):
        common = ErrorCleaner._cleanup_errors(error_message)
        final_message = ErrorCleaner._join_final(common)
        return final_message



# -*- coding: utf-8 -*-
"""
Test xls2xform module.
"""
# The Django application xls2xform uses the function
# pyxform.create_survey. We have a test here to make sure no one
# breaks that function.

import argparse
import logging
from unittest import TestCase

import pyxform
from pyxform.xls2xform import (
    _create_parser,
    _validator_args_logic,
    get_xml_path,
    main_cli,
)

from pyxform.tests.utils import path_to_text_fixture

try:
    from unittest import mock
except ImportError:
    import mock


class XLS2XFormTests(TestCase):
    survey_package = {
        "id_string": "test_2011_08_29b",
        "name_of_main_section": "gps",
        "sections": {
            "gps": {
                "children": [{"name": "location", "type": "gps"}],
                "name": "gps",
                "type": "survey",
            }
        },
        "title": "test",
    }
    survey = pyxform.create_survey(**survey_package)

    def test_create_parser_without_args(self):
        """Should exit when no args provided."""
        with self.assertRaises(SystemExit):
            _create_parser().parse_args([])

    def test_create_parser_optional_output_path(self):
        """
        Should run fine for a single argument i.e. that is the
        path to the xlsx file path, while the output path is left out
        """
        try:
            _create_parser().parse_args(["/some/path/tofile.xlsx"])
        except SystemExit:
            self.fail()

    def test_create_parser_with_args(self):
        """Should parse the provided arguments."""
        arg_xlsform = "xlsform.xlsx"
        arg_output = "."
        arg_list = [
            "--json",
            "--skip_validate",
            "--pretty_print",
            arg_xlsform,
            arg_output,
        ]
        args = _create_parser().parse_args(arg_list)
        self.assertEqual(arg_xlsform, args.path_to_XLSForm)
        self.assertEqual(arg_output, args.output_path)
        self.assertEqual(True, args.json)
        self.assertEqual(False, args.skip_validate)
        self.assertEqual(True, args.pretty_print)

    def test_create_parser_file_name_with_space(self):
        """Should interpret the path correctly."""
        arg_xlsform = "some/path/my xlsform.xlsx"
        arg_output = "."
        arg_list = [arg_xlsform, arg_output]
        args = _create_parser().parse_args(arg_list)
        self.assertEqual(arg_xlsform, args.path_to_XLSForm)

    def test_create_parser_json_default_false(self):
        """Should have json=False if not specified."""
        arg_xlsform = "xlsform.xlsx"
        arg_output = "."
        arg_list = [arg_xlsform, arg_output]
        args = _create_parser().parse_args(arg_list)
        self.assertEqual(False, args.json)

    def test_create_parser_skip_validate_default_true(self):
        """Should have skip_validate=True if not specified."""
        arg_xlsform = "xlsform.xlsx"
        arg_output = "."
        arg_list = [arg_xlsform, arg_output]
        args = _create_parser().parse_args(arg_list)
        self.assertEqual(True, args.skip_validate)

    def test_create_parser_no_enketo_default_false(self):
        """Should have enketo_validate=False if not specified."""
        arg_xlsform = "xlsform.xlsx"
        arg_output = "."
        arg_list = [arg_xlsform, arg_output]
        args = _create_parser().parse_args(arg_list)
        self.assertEqual(False, args.enketo_validate)

    def test_create_parser_pretty_print_default_False(self):
        """Should have pretty_print=False if not specified."""
        args = _create_parser().parse_args(["xlsform.xlsx", "."])
        self.assertFalse(args.pretty_print)

    def test_validator_args_logic_skip_validate_alone(self):
        """Should deactivate both validators."""
        raw_args = _create_parser().parse_args(["xlsform.xlsx", ".", "--skip_validate"])
        args = _validator_args_logic(args=raw_args)
        self.assertEqual(False, args.odk_validate)
        self.assertEqual(False, args.enketo_validate)

    def test_validator_args_logic_odk_default(self):
        """Should activate ODK only."""
        raw_args = _create_parser().parse_args(["xlsform.xlsx", "."])
        args = _validator_args_logic(args=raw_args)
        self.assertEqual(True, args.odk_validate)
        self.assertEqual(False, args.enketo_validate)

    def test_validator_args_logic_enketo_only(self):
        """Should activate Enketo only."""
        raw_args = _create_parser().parse_args(
            ["xlsform.xlsx", ".", "--enketo_validate"]
        )
        args = _validator_args_logic(args=raw_args)
        self.assertEqual(False, args.odk_validate)
        self.assertEqual(True, args.enketo_validate)

    def test_validator_args_logic_odk_only(self):
        """Should activate ODK only."""
        raw_args = _create_parser().parse_args(["xlsform.xlsx", ".", "--odk_validate"])
        args = _validator_args_logic(args=raw_args)
        self.assertEqual(True, args.odk_validate)
        self.assertEqual(False, args.enketo_validate)

    def test_validator_args_logic_odk_and_enketo(self):
        """Should activate ODK and Enketo."""
        raw_args = _create_parser().parse_args(
            ["xlsform.xlsx", ".", "--odk_validate", "--enketo_validate"]
        )
        args = _validator_args_logic(args=raw_args)
        self.assertEqual(True, args.odk_validate)
        self.assertEqual(True, args.enketo_validate)

    def test_validator_args_logic_skip_validate_override(self):
        """Should deactivate both validators"""
        raw_args = _create_parser().parse_args(
            [
                "xlsform.xlsx",
                ".",
                "--skip_validate",
                "--odk_validate",
                "--enketo_validate",
            ]
        )
        args = _validator_args_logic(args=raw_args)
        self.assertEqual(False, args.odk_validate)
        self.assertEqual(False, args.enketo_validate)

    @mock.patch(
        "argparse.ArgumentParser.parse_args",
        return_value=argparse.Namespace(
            path_to_XLSForm="xlsform.xlsx",
            output_path=None,
            json=False,
            skip_validate=False,
            odk_validate=False,
            enketo_validate=False,
            pretty_print=False,
        ),
    )
    @mock.patch("pyxform.xls2xform.xls2xform_convert")
    def test_xls2form_convert_parameters(self, converter_mock, parser_mock_args):
        """
        Checks that xls2xform_convert is given the right arguments, when the
        output-path is not given
        """
        converter_mock.return_value = "{}"
        main_cli()
        converter_mock.assert_called_once_with(
            xlsform_path="xlsform.xlsx",
            xform_path="xlsform.xml",
            validate=False,
            pretty_print=False,
            enketo=False,
        )

    @mock.patch(
        "argparse.ArgumentParser.parse_args",
        return_value=argparse.Namespace(
            path_to_XLSForm="xlsform.xlsx",
            output_path=None,
            json=True,
            skip_validate=False,
            odk_validate=False,
            enketo_validate=False,
            pretty_print=False,
        ),
    )
    @mock.patch("pyxform.xls2xform.xls2xform_convert")
    def test_xls2xform_convert_params_with_flags(
        self, converter_mock, parser_mock_args
    ):
        """
        Should call xlsform_convert with the correct input for output
        path where only the xlsform input path and json flag were provided, since
        the xlsform-convert can be called if json flag was set or when not
        """
        converter_mock.return_value = "{}"
        main_cli()
        converter_mock.assert_called_once_with(
            xlsform_path="xlsform.xlsx",
            xform_path="xlsform.xml",
            validate=False,
            pretty_print=False,
            enketo=False,
        )

    @mock.patch(
        "argparse.ArgumentParser.parse_args",
        return_value=argparse.Namespace(
            path_to_XLSForm=path_to_text_fixture("bad_calc.xlsx"),
            output_path=None,
            json=False,
            skip_validate=True,
            odk_validate=True,
            enketo_validate=True,
            pretty_print=True,
        ),
    )
    def test_xls2xform_convert_throwing_odk_error(self, parser_mock_args):
        """
        Parse and validate bad_calc.xlsx
        """
        logger = logging.getLogger("pyxform.xls2xform")
        with mock.patch.object(logger, "error") as mock_debug:
            main_cli()
            self.assertEqual(mock_debug.call_count, 1)

    def test_get_xml_path_function(self):
        """Should return an xml path in the same directory as the xlsx file"""
        xlsx_path = "/home/user/Desktop/xlsform.xlsx"
        expected = "/home/user/Desktop/xlsform.xml"
        assert expected == get_xml_path(xlsx_path)
        # check that it also handles spaced routes
        xlsx_path = "/home/user/Desktop/my xlsform.xlsx"
        expected = "/home/user/Desktop/my xlsform.xml"
        assert expected == get_xml_path(xlsx_path)



# -*- coding: utf-8 -*-
"""
Test table list appearance syntax.
"""
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase

MD = '''
| survey  |                    |            |           |            |                   |
|         | type               | name       | label     | hint       |appearance         |
|         | begin_group        | tablelist1 | Table_Y_N |            |table-list minimal |
|         | select_one yes_no  | options1a  | Q1        | first row! |                   |
|         | select_one yes_no  | options1b  | Q2        |            |                   |
|         | end_group          |            |           |            |                   |
| choices |                    |            |           |            |                   |
|         | list_name          | name       | label     |            |                   |
|         | yes_no             | yes        | Yes       |            |                   |
0            """  # noqa
'''  # nopep8

XML_CONTAINS = """
    <group appearance="field-list minimal" ref="/table-list-appearance-mod/tablelist1">
      <input ref="/table-list-appearance-mod/tablelist1/generated_table_list_label_2">
        <label>Table_Y_N</label>
      </input>
      <select1 appearance="label" ref="/table-list-appearance-mod/tablelist1/reserved_name_for_field_list_labels_3">
        <label> </label>
        <item>
          <label>Yes</label>
          <value>yes</value>
        </item>
      </select1>
      <select1 appearance="list-nolabel" ref="/table-list-appearance-mod/tablelist1/options1a">
        <label>Q1</label>
        <hint>first row!</hint>
""".strip()  # nopep8


class TableListTest(PyxformTestCase):
    def test_table_list(self):
        self.assertPyxformXform(
            name="table-list-appearance-mod",
            md=MD,
            xml__contains=[XML_CONTAINS],
            debug=False,
        )



# -*- coding: utf-8 -*-
"""
The last-saved virtual instance can be queried to get values from the last saved instance of the form being authored.
"""
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class LastSavedTest(PyxformTestCase):
    def test_last_saved_in_calculate(self):
        self.assertPyxformXform(
            name="last-saved",
            md="""
            | survey |            |          |       |                                    |
            |        | type       | name     | label | calculation                        |
            |        | text       | foo      | Foo   |                                    |
            |        | calculate  | last-foo |       | concat("Foo: ", ${last-saved#foo}) |
            """,
            xml__contains=[
                '<instance id="__last-saved" src="jr://instance/last-saved"/>',
                "calculate=\"concat(&quot;Foo: &quot;,  instance('__last-saved')/last-saved/foo )\"",
            ],
        )

    def test_last_saved_calculate_self(self):
        self.assertPyxformXform(
            name="last-saved",
            md="""
            | survey |            |          |       |                       |
            |        | type       | name     | label | calculation           |
            |        | integer    | foo      | Foo   | ${last-saved#foo} + 1 |
            """,
            xml__contains=[
                '<instance id="__last-saved" src="jr://instance/last-saved"/>',
                'calculate=" instance(\'__last-saved\')/last-saved/foo  + 1" nodeset="/last-saved/foo"',
            ],
        )

    def test_last_saved_in_label(self):
        self.assertPyxformXform(
            name="last-saved",
            md="""
            | survey |            |          |                                      |
            |        | type       | name     | label                                |
            |        | text       | foo      | Foo was previously ${last-saved#foo} |
            """,
            xml__contains=[
                "<label> Foo was previously <output value=\" instance('__last-saved')/last-saved/foo \"/> </label>",
            ],
        )

    def test_multiple_last_saved_in_calculate(self):
        self.assertPyxformXform(
            name="last-saved",
            md="""
            | survey |            |          |       |                                       |
            |        | type       | name     | label | calculation                           |
            |        | integer    | foo      | Foo   |                                       |
            |        | integer    | bar      | Bar   |                                       |
            |        | calculate  | last-sum |       | ${last-saved#foo} + ${last-saved#bar} |
            """,
            xml__contains=[
                '<instance id="__last-saved" src="jr://instance/last-saved"/>',
                "calculate=\" instance('__last-saved')/last-saved/foo  +  instance('__last-saved')/last-saved/bar \"",
            ],
        )

    def test_last_saved_in_relevant(self):
        self.assertPyxformXform(
            name="last-saved",
            md="""
            | survey |            |          |       |                        |
            |        | type       | name     | label | relevant               |
            |        | integer    | foo      | Foo   | ${last-saved#foo} < 12 |
            """,
            xml__contains=[
                '<instance id="__last-saved" src="jr://instance/last-saved"/>',
                "relevant=\" instance('__last-saved')/last-saved/foo  &lt; 12\"",
            ],
        )

    def test_last_saved_in_readonly(self):
        self.assertPyxformXform(
            name="last-saved",
            md="""
            | survey |            |          |       |                        |
            |        | type       | name     | label | readonly               |
            |        | integer    | foo      | Foo   | ${last-saved#foo} < 12 |
            """,
            xml__contains=[
                '<instance id="__last-saved" src="jr://instance/last-saved"/>',
                "readonly=\" instance('__last-saved')/last-saved/foo  &lt; 12\"",
            ],
        )

    def test_last_saved_in_constraint(self):
        self.assertPyxformXform(
            name="last-saved",
            md="""
            | survey |            |          |       |                        |
            |        | type       | name     | label | constraint             |
            |        | integer    | foo      | Foo   | . > ${last-saved#foo}  |
            """,
            xml__contains=[
                '<instance id="__last-saved" src="jr://instance/last-saved"/>',
                "constraint=\". &gt;  instance('__last-saved')/last-saved/foo \"",
            ],
        )

    def test_last_saved_in_required(self):
        self.assertPyxformXform(
            name="last-saved",
            md="""
            | survey |            |          |       |                        |
            |        | type       | name     | label | required               |
            |        | integer    | foo      | Foo   | ${last-saved#foo} < 12 |
            """,
            xml__contains=[
                '<instance id="__last-saved" src="jr://instance/last-saved"/>',
                "required=\" instance('__last-saved')/last-saved/foo  &lt; 12\"",
            ],
        )

    def test_last_saved_in_default(self):
        self.assertPyxformXform(
            name="last-saved",
            md="""
            | survey |            |          |       |                   |
            |        | type       | name     | label | default           |
            |        | integer    | foo      | Foo   | ${last-saved#foo} |
            """,
            xml__contains=[
                '<instance id="__last-saved" src="jr://instance/last-saved"/>',
                '<setvalue event="odk-instance-first-load" ref="/last-saved/foo" value=" instance(\'__last-saved\')/last-saved/foo "/>',
            ],
        )

    # This is not really a useful thing to do because we get an unqualified reference. Clients should use the first
    # node in the repeat in an expression like this one where a single value is needed or the full nodeset in an
    # expression where a nodeset is required. An alternative to consider is to generate an expression with position(..).
    def test_last_saved_in_repeat(self):
        self.assertPyxformXform(
            name="last-saved",
            md="""
            | survey |              |           |       |                            |
            |        | type         | name      | label | calculation                |
            |        | begin repeat | my-repeat |       |                            |
            |        | integer      | foo       | Foo   |                            |
            |        | calculate    | bar       |       | ${foo} + ${last-saved#foo} |
            |        | end repeat   | my-repeat |       |                            |
            |        | calculate    | baz       |       | ${foo} + ${last-saved#foo} |
            """,
            run_odk_validate=False,  # This test is more for documentation. The unqualified ref fails Validate.
            xml__contains=[
                '<instance id="__last-saved" src="jr://instance/last-saved"/>',
                'calculate=" ../foo  +  instance(\'__last-saved\')/last-saved/my-repeat/foo " nodeset="/last-saved/my-repeat/bar"',
                'calculate=" /last-saved/my-repeat/foo  +  instance(\'__last-saved\')/last-saved/my-repeat/foo " nodeset="/last-saved/baz"',
            ],
        )

    # In repeats, ${} references in choice filters get expanded to relative references using current(). ${last-saved}
    # references should be expanded to absolute references.
    def test_last_saved_in_choice_filter_does_not_use_current_or_relative_ref(self):
        self.assertPyxformXform(
            name="last-saved",
            md="""
            | survey |                |          |       |                                        |
            |        | type           | name     | label | choice_filter                          |
            |        | begin repeat   | repeat   |       |                                        |
            |        | select_one foo | foo      | Foo   | not(selected(${last-saved#foo}, name)) |
            |        | select_one foo | bar      | Bar   | not(selected(${foo}, name))            |
            |        | end repeat     | repeat   |       |                                        |
            | choices|                |          |       |                                        |
            |        | list_name      | name     | label |                                        |
            |        | foo            | a        | A     |                                        |
            """,
            xml__contains=[
                '<instance id="__last-saved" src="jr://instance/last-saved"/>',
                "<itemset nodeset=\"instance('foo')/root/item[not(selected( instance('__last-saved')/last-saved/repeat/foo , name))]\">",
                "<itemset nodeset=\"instance('foo')/root/item[not(selected( current()/../foo , name))]\">",
            ],
        )

    # Last saved must refer to an instance of this form definition so we should not allow fields that don't exist in
    # this definition.
    def test_last_saved_errors_when_field_does_not_exist(self):
        self.assertPyxformXform(
            name="last-saved",
            md="""
            | survey |            |          |       |                                   |
            |        | type       | name     | label | calculation                       |
            |        | calculate  | last-foo |       | concat("Foo: ", ${last-saved#foo} |
            """,
            errored=True,
            error__contains=[
                "There has been a problem trying to replace ${last-saved#foo} with the XPath to the survey element named 'foo'. There is no survey element with this name."
            ],
        )

    def test_last_saved_not_generated_without_last_save_call(self):
        self.assertPyxformXform(
            name="last-saved",
            md="""
            | survey |            |          |       |             |
            |        | type       | name     | label | calculation |
            |        | integer    | foo      | Foo   |             |
            |        | calculate  | bar      | Bar   | ${foo} + 4  |
            """,
            xml__excludes=[
                '<instance id="__last-saved" src="jr://instance/last-saved"/>'
            ],
        )

    def test_last_saved_call_and___last_saved_instance_conflict(self):
        self.assertPyxformXform(
            name="last-saved",
            md="""
            | survey |              |              |       |                       |
            |        | type         | name         | label | calculation           |
            |        | xml-external | __last-saved |       |                       |
            |        | integer      | foo          | Foo   |                       |
            |        | calculate    | bar          | Bar   | ${last-saved#foo} + 4 |
            """,
            errored=True,
            error__contains=[
                "The same instance id will be generated for different external instance source URIs. Please check the form.",
                "Instance name: '__last-saved', Existing type: 'external', Existing URI: 'jr://file/__last-saved.xml'",
                "Duplicate type: 'instance', Duplicate URI: 'jr://instance/last-saved', Duplicate context: 'None'.",
            ],
        )



# -*- coding: utf-8 -*-
"""
Test setting form name to data.
"""
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase
from pyxform.utils import unicode


class FormNameTest(PyxformTestCase):
    def test_default_to_data_when_no_name(self):
        """
        Test no form_name will default to survey name to 'data'.
        """
        survey = self.md_to_pyxform_survey(
            """
            | survey   |           |      |           |
            |          | type      | name | label     |
            |          | text      | city | City Name |
            | settings |           |      |           |
            |          | id_string | name |
            |          | some-id   | data |
            """,
            kwargs={},
            autoname=False,
        )

        # We're passing autoname false when creating the survey object.
        self.assertEqual(survey.id_string, None)
        self.assertEqual(survey.name, unicode("data"))
        self.assertEqual(survey.title, None)

        # Set required fields because we need them if we want to do xml comparison.
        survey.id_string = "some-id"
        survey.title = "data"

        self.assertPyxformXform(
            survey=survey,
            instance__contains=['<data id="some-id">'],
            model__contains=['<bind nodeset="/data/city" type="string"/>'],
            xml__contains=[
                '<input ref="/data/city">',
                "<label>City Name</label>",
                "</input>",
            ],
        )

    def test_default_to_data(self):
        """
        Test using data as the name of the form which will generate <data />.
        """
        self.assertPyxformXform(
            md="""
               | survey |      |      |           |
               |        | type | name | label     |
               |        | text | city | City Name |
               """,
            name="data",
            id_string="some-id",
            instance__contains=['<data id="some-id">'],
            model__contains=['<bind nodeset="/data/city" type="string"/>'],
            xml__contains=[
                '<input ref="/data/city">',
                "<label>City Name</label>",
                "</input>",
            ],
        )

    def test_default_form_name_to_superclass_definition(self):
        """
        Test no form_name and setting name field, should use name field.
        """

        self.assertPyxformXform(
            md="""
               | survey |      |      |           |
               |        | type | name | label     |
               |        | text | city | City Name |
               """,
            name="some-name",
            id_string="some-id",
            instance__contains=['<some-name id="some-id">'],
            model__contains=['<bind nodeset="/some-name/city" type="string"/>'],
            xml__contains=[
                '<input ref="/some-name/city">',
                "<label>City Name</label>",
                "</input>",
            ],
        )



# -*- coding: utf-8 -*-
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class ChoicesSheetTest(PyxformTestCase):
    def test_numeric_choice_names__for_static_selects__allowed(self):
        """
        Test numeric choice names for static selects.
        """
        self.assertPyxformXform(
            md="""
            | survey   |                    |      |       |
            |          | type               | name | label |
            |          | select_one choices | a    | A     |
            | choices  |                    |      |       |
            |          | list_name          | name | label |
            |          | choices            | 1    | One   |
            |          | choices            | 2    | Two   |
            """,
            xml__contains=["<value>1</value>"],
        )

    def test_numeric_choice_names__for_dynamic_selects__allowed(self):
        """
        Test numeric choice names for dynamic selects.
        """
        self.assertPyxformXform(
            md="""
            | survey   |                    |      |       |               |
            |          | type               | name | label | choice_filter |
            |          | select_one choices | a    | A     | true()        |
            | choices  |                    |      |       |
            |          | list_name          | name | label |
            |          | choices            | 1    | One   |
            |          | choices            | 2    | Two   |
            """,
            xml__contains=['<instance id="choices">', "<item>", "<name>1</name>"],
        )

    def test_choices_without_labels__for_static_selects__allowed(self):
        """
        Test choices without labels for static selects. Validate will NOT fail.
        """
        self.assertPyxformXform(
            md="""
            | survey   |                    |      |       |
            |          | type               | name | label |
            |          | select_one choices | a    | A     |
            | choices  |                    |      |       |
            |          | list_name          | name | label |
            |          | choices            | 1    |       |
            |          | choices            | 2    |       |
            """,
            xml__contains=["<value>1</value>"],
        )

    def test_choices_without_labels__for_dynamic_selects__allowed_by_pyxform(self):
        """
        Test choices without labels for dynamic selects. Validate will fail.
        """
        self.assertPyxformXform(
            md="""
            | survey   |                    |      |       |               |
            |          | type               | name | label | choice_filter |
            |          | select_one choices | a    | A     | true()        |
            | choices  |                    |      |       |
            |          | list_name          | name | label |
            |          | choices            | 1    |       |
            |          | choices            | 2    |       |
            """,
            run_odk_validate=False,
            xml__contains=['<instance id="choices">', "<item>", "<name>1</name>"],
        )



# -*- coding: utf-8 -*-
"""
The validators utility functions.
"""
import collections
import io
import logging
import os
import signal
import subprocess
import tempfile
import threading
import time
from contextlib import closing
from subprocess import PIPE, Popen

from pyxform.errors import PyXFormError

try:
    from urllib.request import urlopen, Request
    from urllib.error import URLError, HTTPError
except ImportError:
    from urllib2 import urlopen, Request, URLError, HTTPError


HERE = os.path.abspath(os.path.dirname(__file__))
XFORM_SPEC_PATH = os.path.join(
    os.path.dirname(HERE), "tests", "test_expected_output", "xlsform_spec_test.xml"
)


class PopenResult:
    """Result data for run_popen_with_timeout"""

    def __init__(
        self, return_code: int, timeout: bool, stdout: bytes, stderr: bytes
    ) -> None:
        self.return_code: int = return_code
        self.timeout: bool = timeout
        self.stdout: str = decode_stream(stream=stdout)
        self.stderr: str = decode_stream(stream=stderr)


# Adapted from:
# http://betabug.ch/blogs/ch-athens/1093
def run_popen_with_timeout(command, timeout) -> "PopenResult":
    """
    Run a sub-program in subprocess.Popen, pass it the input_data,
    kill it if the specified timeout has passed.
    returns a tuple of resultcode, timeout, stdout, stderr
    """
    kill_check = threading.Event()

    def _kill_process_after_a_timeout(pid):
        os.kill(pid, signal.SIGTERM)
        kill_check.set()  # tell the main routine that we had to kill
        # use SIGKILL if hard to kill...
        return

    startup_info = None
    env = None
    if os.name == "nt":
        # Workarounds for pyinstaller
        # https://github.com/pyinstaller/pyinstaller/wiki/Recipe-subprocess
        # disable command window when run from pyinstaller
        startup_info = subprocess.STARTUPINFO()
        # Less fancy version of bitwise-or-assignment (x |= y) shown in ref url.
        if startup_info.dwFlags == 1 or subprocess.STARTF_USESHOWWINDOW == 1:
            startup_info.dwFlags = 1
        else:
            startup_info.dwFlags = 0

        # Workaround for Java sometimes not being able to use the temp directory.
        # https://docs.oracle.com/javase/8/docs/api/java/io/File.html
        # CreateTempFile refers to "java.io.tmpdir" which refers to env vars.
        env = {
            k: v if v is not None else tempfile.gettempdir()
            for k, v in {
                k: os.environ.get(k) for k in ("TEMP", "TMP", "TMPDIR")
            }.items()
        }

    p = Popen(
        command, env=env, stdin=PIPE, stdout=PIPE, stderr=PIPE, startupinfo=startup_info
    )
    watchdog = threading.Timer(timeout, _kill_process_after_a_timeout, args=(p.pid,))
    watchdog.start()
    (stdout, stderr) = p.communicate()
    watchdog.cancel()  # if it's still waiting to run
    timeout = kill_check.isSet()
    kill_check.clear()
    return PopenResult(
        return_code=p.returncode, timeout=timeout, stdout=stdout, stderr=stderr
    )


def decode_stream(stream):
    """
    Decode a stream, e.g. stdout or stderr.

    On Windows, stderr may be latin-1; in which case utf-8 decode will fail.
    If both utf-8 and latin-1 decoding fail then raise all as IOError.
    If the above validate jar call fails, add make sure that the java path
    is set, e.g. PATH=C:\\Program Files (x86)\\Java\\jre1.8.0_71\\bin
    """
    try:
        return stream.decode("utf-8")
    except UnicodeDecodeError as ude:
        try:
            return stream.decode("latin-1")
        except BaseException as be:
            msg = "Failed to decode validate stderr as utf-8 or latin-1."
            raise IOError(msg, ude, be)


def request_get(url):
    """
    Get the response content from URL.
    """
    try:
        r = Request(url)
        r.add_header("Accept", "application/json")
        with closing(urlopen(r)) as u:
            content = u.read()
        if len(content) == 0:
            raise PyXFormError("Empty response from URL: '{u}'.".format(u=url))
        else:
            return content
    except HTTPError as e:
        raise PyXFormError(
            "Unable to fulfill request. Error code: '{c}'. "
            "Reason: '{r}'. URL: '{u}'."
            "".format(r=e.reason, c=e.code, u=url)
        )
    except URLError as e:
        raise PyXFormError(
            "Unable to reach a server. Reason: {r}. "
            "URL: {u}".format(r=e.reason, u=url)
        )


class CapturingHandler(logging.Handler):
    """
    A logging handler capturing all (raw and formatted) logging output.

    Similar concept to "from unittest.case import _CapturingHandler". However
    the watcher.output is a dict keyed by the log level name, instead of a list
    of messages. The watcher.records list has the full LogRecord instances
    which carry extra data about the log source and context.

    Usage:
    Here's how to attach it to a logger, get some logs, and then detach it.

    my_logger = logging.getLogger(name="logger_name")
    capture_handler = CapturingHandler(logger=my_logger)
    info_log_messages = capture_handler.watcher.output["INFO"]
    log_records = capture_handler.watcher.records
    my_logger.removeHandler(hdlr=capture_handler)
    """

    def __init__(self, logger):
        logging.Handler.__init__(self)
        self.watcher = self._get_watcher()
        logger.addHandler(self)
        logger.propagate = False
        logger.setLevel("INFO")
        self.setFormatter(logging.Formatter("%(message)s"))

    def flush(self):
        pass

    def emit(self, record):
        self.watcher.records.append(record)
        msg = self.format(record)
        self.watcher.output[record.levelname].append(msg)

    @staticmethod
    def _get_watcher():
        _LoggingWatcher = collections.namedtuple(
            "_LoggingWatcher", ["records", "output"]
        )
        levels = ["CRITICAL", "ERROR", "WARNING", "INFO", "DEBUG"]
        return _LoggingWatcher([], {x: [] for x in levels})

    def reset(self):
        self.watcher = self._get_watcher()


def check_readable(file_path, retry_limit=10, wait_seconds=0.5):
    """
    Check if a file is readable: True if so, IOError if not. Retry as per args.

    If a file that needs to be read may be locked by some other process (such
    as for reading or writing), this can help avoid an error by waiting for the
    lock to clear.

    :param file_path: Path to file to check.
    :param retry_limit: Number of attempts to read the file.
    :param wait_seconds: Amount of sleep time between read attempts.
    :return: True or raise IOError.
    """

    def catch_try():
        try:
            with io.open(file_path, mode="r"):
                return True
        except IOError:
            return False

    tries = 0
    while not catch_try():
        if tries < retry_limit:
            tries += 1
            time.sleep(wait_seconds)
        else:
            raise IOError("Could not read file: {f}".format(f=file_path))
    return True



# -*- coding: utf-8 -*-
"""
Test translations syntax.
"""
from __future__ import unicode_literals

from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class DoubleColonTranslations(PyxformTestCase):
    def test_langs(self):
        model_contains = (
            """<bind nodeset="/translations/n1"""
            + """" readonly="true()" type="string"/>"""
        )
        self.assertPyxformXform(
            name="translations",
            id_string="transl",
            md="""
            | survey |      |      |                |               |
            |        | type | name | label::english | label::french |
            |        | note | n1   | hello          | bonjour       |
            """,
            errored=False,
            itext__contains=[
                '<translation lang="french">',
                '<text id="/translations/n1:label">',
                "<value>bonjour</value>",
                "</text>",
                "</translation>",
                '<translation lang="english">',
                '<text id="/translations/n1:label">',
                "<value>hello</value>",
                "</text>",
                "</translation>",
            ],
            xml__contains=["""<label ref="jr:itext('/translations/n1:label')"/>"""],
            model__contains=[model_contains],
        )


class TransaltionsTest(PyxformTestCase):
    """Test XLSForm translations."""

    def test_missing_media_itext(self):
        """Test missing media itext translation

        Fix for https://github.com/XLSForm/pyxform/issues/32
        """
        xform_md = """
        | survey |                                |                    |                                                              |                                               |                 |          |                      |                       |
        |        | name                           | type               | label                                                        | label::Russian                                | label::Kyrgyz   | required | media::audio::Kyrgyz | media::audio::Russian |
        |        | Have_you_received_informed_con | select_one bt7nj36 | A.01 Have you received informed consent from the respondent? | Получили ли вы форму согласия от респондента? | This is Kyrgyz. | true     | something.mp3        | test.mp3              |
        | choices |           |      |       |                |               |
        |         | list name | name | label | label::Russian | label::Kyrgyz |
        |         | bt7nj36   | 0    | No    | Нет            | Нет (ky)      |
        |         | bt7nj36   | 1    | Yes   | Да             | Да (ky)       |
        """
        self.assertPyxformXform(
            name="multi_language_form",
            id_string="multi_language_form",
            md=xform_md,
            errored=False,
            debug=False,
            itext__contains=[
                '<translation default="true()" lang="default">',
                '<text id="/multi_language_form/Have_you_received_informed_con:label">',
                "A.01 Have you received informed consent from the respondent?",
                '<translation lang="Russian">',
                "<value>Получили ли вы форму согласия от респондента?</value>",
                '<value form="audio">jr://audio/test.mp3</value>',
                '<translation lang="Kyrgyz">',
                "<value>This is Kyrgyz.</value>",
                '<value form="audio">jr://audio/something.mp3</value>',
            ],
            itext__excludes=['<value form="audio">jr://audio/-</value>'],
        )



# -*- coding: utf-8 -*-
"""
Test XLSForm sheet names.
"""
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase
from pyxform.tests.utils import prep_for_xml_contains


class InvalidSurveyColumnsTests(PyxformTestCase):
    """
    Invalid survey column tests
    """

    def test_missing_name(self):
        """
        every question needs a name (or alias of name)
        """
        self.assertPyxformXform(
            name="invalidcols",
            ss_structure={"survey": [{"type": "text", "label": "label"}]},
            errored=True,
            error__contains=["no name"],
        )

    def test_missing_name_but_has_alias_of_name(self):
        self.assertPyxformXform(
            name="invalidcols",
            ss_structure={
                "survey": [{"value": "q1", "type": "text", "label": "label"}]
            },
            errored=False,
        )

    def test_missing_label(self):
        self.assertPyxformXform(
            name="invalidcols",
            ss_structure={"survey": [{"type": "text", "name": "q1"}]},
            errored=True,
            error__contains=["no label or hint"],
        )

    def test_label_node_added_when_hint_given_but_no_label_value(self):
        """Should output a label node even if no label is specified."""
        expected = """
          <input ref="/data/a">
            <label/>
            <hint>h</hint>
          </input>
          <input ref="/data/b">
            <label>l</label>
          </input>
          <input ref="/data/c">
            <label ref="jr:itext('/data/c:label')"/>
          </input>
        """
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |      |      |       |              |      |
            |        | type | name | label | media::image | hint |
            |        | text | a    |       |              | h    |
            |        | text | b    | l     |              |      |
            |        | text | c    |       | m.png        |      |
            """,
            xml__contains=prep_for_xml_contains(expected),
        )

    def test_column_case(self):
        """
        Ensure that column name is case insensitive
        """
        self.assertPyxformXform(
            name="mixedcasecolumns",
            md="""
            | Survey |         |         |               |
            |        | Type    | name    | Label         |
            |        | text    | Name    | the name      |
            |        | integer | age     | the age       |
            |        | text    | gender  | the gender    |
            """,
            errored=False,
            debug=False,
        )


class InvalidChoiceSheetColumnsTests(PyxformTestCase):
    """
    Invalid choice sheet column tests
    """

    def _simple_choice_ss(self, choice_sheet=None):
        """
        Return simple choices sheet
        """

        if choice_sheet is None:
            choice_sheet = []
        return {
            "survey": [
                {
                    "type": "select_one l1",
                    "name": "l1choice",
                    "label": "select one from list l1",
                }
            ],
            "choices": choice_sheet,
        }

    def test_valid_choices_sheet_passes(self):
        """
        Test invalid choices sheet passes
        """

        self.assertPyxformXform(
            name="valid_choices",
            ss_structure=self._simple_choice_ss(
                [
                    {"list_name": "l1", "name": "c1", "label": "choice 1"},
                    {"list_name": "l1", "name": "c2", "label": "choice 2"},
                ]
            ),
            errored=False,
        )

    def test_invalid_choices_sheet_fails(self):
        """
        Test invalid choices sheet fails
        """

        self.assertPyxformXform(
            name="missing_name",
            ss_structure=self._simple_choice_ss(
                [
                    {"list_name": "l1", "label": "choice 1"},
                    {"list_name": "l1", "label": "choice 2"},
                ]
            ),
            errored=True,
            error__contains=["option with no name"],
        )

    def test_missing_list_name(self):
        """
        Test missing sheet name
        """

        self.assertPyxformXform(
            name="missing_list_name",
            ss_structure=self._simple_choice_ss(
                [
                    {"bad_column": "l1", "name": "l1c1", "label": "choice 1"},
                    {"bad_column": "l1", "name": "l1c1", "label": "choice 2"},
                ]
            ),
            debug=False,
            errored=True,
            # some basic keywords that should be in the error:
            error__contains=["choices", "name", "list_name"],
        )

    def test_clear_filename_error_message(self):
        """Test clear filename error message"""
        error_message = (
            "The name 'bad@filename' is an invalid XML tag, it "
            "contains an invalid character '@'. Names must begin"
            " with a letter, colon, or underscore, subsequent "
            "characters can include numbers, dashes, and periods"
        )
        self.assertPyxformXform(
            name="bad@filename",
            ss_structure=self._simple_choice_ss(
                [
                    {"list_name": "l1", "name": "c1", "label": "choice 1"},
                    {"list_name": "l1", "name": "c2", "label": "choice 2"},
                ]
            ),
            errored=True,
            error__contains=[error_message],
        )


class AliasesTests(PyxformTestCase):
    """
    Aliases Tests
    """

    def test_value_and_name(self):
        """
        confirm that both 'name' and 'value' columns of choice list work
        """
        for name_alias in ["name", "value"]:
            self.assertPyxformXform(
                name="aliases",
                md="""
                | survey  |               |                |            |
                |         | type          | name           | label      |
                |         | select_one yn | q1             | Question 1 |
                | choices |               |                |            |
                |         | list name     | %(name_alias)s | label      |
                |         | yn            | yes            | Yes        |
                |         | yn            | no             | No         |
                """
                % ({"name_alias": name_alias}),
                instance__contains=["<q1/>"],
                model__contains=['<bind nodeset="/aliases/q1" type="string"/>'],
                xml__contains=[
                    '<select1 ref="/aliases/q1">',
                    "<value>yes</value>",
                    "<value>no</value>",
                    "</select1>",
                ],
            )


''' # uncomment when re-implemented
    # TODO: test that this fails for the correct reason
    def test_conflicting_aliased_values_raises_error(self):
        # example:
        # an xlsform has {"name": "q_name", "value": "q_value"}
        # should not compile because "name" and "value" columns are aliases

        self.assertPyxformXform(
            # debug=True,
            name="aliases",
            md="""
            | survey |      |        |         |            |
            |        | type | name   | value   | label      |
            |        | text | q_name | q_value | Question 1 |
            """,
            errored=True,
        )
'''



# -*- coding: utf-8 -*-
"""
Test loop question type.
"""
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class TestLoop(PyxformTestCase):
    """
    Test loop question type.
    """

    def test_loop(self):
        """
        Test loop question type.
        """
        self.assertPyxformXform(
            name="test_loop",
            md="""
                | survey |              |           |                |        |
                |        |     type     |    name   | bind:relevant  | label  |
                |        | begin repeat | for-block |                | Oh HAI |
                |        | string       | input     | (${done}='no') | HI HI  |
                |        | string       | done      |                | DONE?  |
                |        | end repeat   |           |                |        |
                """,
            instance__contains=['<for-block jr:template="">', "</for-block>"],
            model__contains=[
                """<bind nodeset="/test_loop/for-block/input" """
                """relevant="( ../done ='no')" """
                """type="string"/>"""
            ],
            xml__contains=[
                '<group ref="/test_loop/for-block">',
                "<label>Oh HAI</label>",
                "</group>",
            ],
        )



# -*- coding: utf-8 -*-
"""
Guidance hint test module.
"""
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class GuidanceHintTest(PyxformTestCase):
    """Test guidance_hint XLSForms."""

    def test_hint_only(self):
        """Test hint only column."""
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |        |          |       |           |
            |        | type   |   name   | label | hint      |
            |        | string |   name   | Name  | your name |
            """,
            xml__contains=["<hint>your name</hint>"],
        )

    def test_guidance_hint_and_label(self):
        """Test guidance_hint with label"""
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |        |          |       |                              |
            |        | type   |   name   | label | guidance_hint                |
            |        | string |   name   | Name  | as shown on birth certificate|
            """,  # noqa
            xml__contains=[
                "<hint ref=\"jr:itext('/data/name:hint')\"/>",
                '<value form="guidance">as shown on birth certificate</value>',
                "<hint ref=\"jr:itext('/data/name:hint')\"/>",
            ],
            run_odk_validate=True,
        )

    def test_hint_and_guidance_one_language(self):  # pylint: disable=C0103
        """Test guidance_hint in one language."""
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |        |          |       |           |                              |
            |        | type   |   name   | label | hint      | guidance_hint                |
            |        | string |   name   | Name  | your name | as shown on birth certificate|
            """,  # noqa
            xml__contains=[
                "<hint ref=\"jr:itext('/data/name:hint')\"/>",
                "<value>your name</value>",
                '<value form="guidance">as shown on birth certificate</value>',
            ],
        )

    def test_multi_language_guidance(self):
        """Test guidance_hint in multiple languages."""
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |        |          |       |           |                              |                                     |
            |        | type   |   name   | label | hint      | guidance_hint                | guidance_hint::French (fr)          |
            |        | string |   name   | Name  | your name | as shown on birth certificate| comme sur le certificat de naissance|
            """,  # noqa
            xml__contains=[
                '<translation lang="French (fr)">',
                '<value form="guidance">comme sur le certificat de naissance</value>',  # noqa
                '<translation default="true()" lang="default">',
                '<value form="guidance">as shown on birth certificate</value>',
                "<hint ref=\"jr:itext('/data/name:hint')\"/>",
            ],
        )

    def test_guidance_hint_only(self):
        """Test guidance_hint only."""
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |        |      |                               |
            |        | type   | name | guidance_hint                 |
            |        | string | name | as shown on birth certificate |
            """,
            errored=True,
            error__contains=["The survey element named 'name' has no label or hint."],
        )

    def test_multi_language_guidance_only(self):  # pylint:disable=C0103
        """Test guidance_hint only in multiple languages."""
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |        |      |                              |                                      |
            |        | type   | name | guidance_hint                | guidance_hint::French (fr)           |
            |        | string | name | as shown on birth certificate| comme sur le certificat de naissance |
            """,  # noqa
            errored=True,
            error__contains=["The survey element named 'name' has no label or hint."],
        )

    def test_multi_language_hint(self):
        """Test hint in multiple languages."""
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |        |          |                      |                    |
            |        | type   |   name   | hint                 | hint::French (fr)  |
            |        | string |   name   | default language hint| French hint        |
            """,  # noqa
            xml__contains=[
                "<hint ref=\"jr:itext('/data/name:hint')\"/>",
                "<value>French hint</value>",
                "<value>default language hint</value>",
            ],
        )



# -*- coding: utf-8 -*-
"""
Test setgeopoint widget.
"""
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class SetGeopointTest(PyxformTestCase):
    """Test setgeopoint widget class."""

    def test_setgeopoint(self):
        self.assertPyxformXform(
            name="data",
            md="""
            | survey |                |             |          |
            |        | type           | name        | label    |
            |        | start-geopoint | my-location | my label |
            """,
            xml__contains=[
                '<bind nodeset="/data/my-location" type="geopoint"/>',
                '<odk:setgeopoint event="odk-instance-first-load" ref="/data/my-location"/>',
                "",
            ],
            run_odk_validate=True,
        )



# -*- coding: utf-8 -*-
"""
Test missing headers in XLSForm.
"""
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class MissingHeaders(PyxformTestCase):
    """Test missing headers in XLSForm

    When survey and choices columns are missing headers, it is helpful to see
    an error message that prompts the user to include necessary headers.
    """

    def test_missing_survey_headers(self):
        self.assertPyxformXform(
            md="""
            | survey |                 |    |
            |        | select_one list | S1 |
            """,
            errored=True,
            error__contains=["missing important column headers"],
        )

    def test_missing_choice_headers(self):
        self.assertPyxformXform(
            md="""
            | survey  |                 |          |      |
            |         | type            | label    | name |
            |         | select_one list | S1       | s1   |
            | choices |                 |          |      |
            |         | list            | option a | a    |
            |         | list            | option b | b    |
            """,
            errored=True,
            error__contains=[
                "has the mandatory columns 'list_name', 'name', and 'label'"
            ],
        )



# -*- coding: utf-8 -*-
"""
Test xml-external syntax and instances generated from pulldata calls.

See also test_external_instances_for_selects
"""
from pyxform.errors import PyXFormError
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase, PyxformTestError
from textwrap import dedent


class ExternalInstanceTests(PyxformTestCase):
    """
    External Instance Tests
    """

    def test_can__output_single_external_xml_item(self):
        """Simplest possible example to include an external instance."""
        self.assertPyxformXform(
            md="""
            | survey |              |        |       |
            |        | type         | name   | label |
            |        | xml-external | mydata |       |
            """,
            model__contains=['<instance id="mydata" src="jr://file/mydata.xml"/>'],
        )

    def test_can__output_single_external_csv_item(self):
        """Simplest possible example to include an external instance."""
        self.assertPyxformXform(
            md="""
            | survey |              |        |       |
            |        | type         | name   | label |
            |        | csv-external | mydata |       |
            """,
            model__contains=['<instance id="mydata" src="jr://file-csv/mydata.csv"/>'],
        )

    def test_cannot__use_same_external_xml_id_in_same_section(self):
        """Duplicate external instances in the same section raises an error."""
        with self.assertRaises(PyxformTestError) as ctx:
            self.assertPyxformXform(
                md="""
                | survey |              |        |       |
                |        | type         | name   | label |
                |        | xml-external | mydata |       |
                |        | xml-external | mydata |       |
                """,
                model__contains=[],
            )
        # This is caught first by existing validation rule.
        self.assertIn(
            "There are more than one survey elements named 'mydata'",
            repr(ctx.exception),
        )

    def test_can__use_unique_external_xml_in_same_section(self):
        """Two unique external instances in the same section is OK."""
        self.assertPyxformXform(
            md="""
            | survey |              |         |       |
            |        | type         | name    | label |
            |        | xml-external | mydata  |       |
            |        | xml-external | mydata2 |       |
            """,
            model__contains=[
                '<instance id="mydata" src="jr://file/mydata.xml"/>',
                '<instance id="mydata2" src="jr://file/mydata2.xml"/>',
            ],
        )

    def test_can__use_unique_external_csv_in_same_section(self):
        """Two unique external instances in the same section is OK."""
        self.assertPyxformXform(
            md="""
            | survey |              |         |       |
            |        | type         | name    | label |
            |        | csv-external | mydata  |       |
            |        | csv-external | mydata2 |       |
            """,
            model__contains=[
                '<instance id="mydata" src="jr://file-csv/mydata.csv"/>',
                '<instance id="mydata2" src="jr://file-csv/mydata2.csv"/>',
            ],
        )

    def test_cannot__use_same_external_xml_id_across_groups(self):
        """Duplicate external instances anywhere raises an error."""
        with self.assertRaises(PyxformTestError) as ctx:
            self.assertPyxformXform(
                md="""
                | survey |              |        |       |
                |        | type         | name   | label |
                |        | xml-external | mydata |       |
                |        | begin group  | g1     |       |
                |        | xml-external | mydata |       |
                |        | end group    | g1     |       |
                |        | begin group  | g2     |       |
                |        | xml-external | mydata |       |
                |        | end group    | g2     |       |
                """,
                model__contains=[],
            )
        self.assertIn("Instance names must be unique", repr(ctx.exception))
        self.assertIn("The name 'mydata' was found 3 time(s)", repr(ctx.exception))

    def test_cannot__use_external_xml_and_csv_with_same_filename(self):
        """Duplicate external instances anywhere raises an error."""
        with self.assertRaises(PyxformTestError) as ctx:
            self.assertPyxformXform(
                md="""
                | survey |              |        |       |
                |        | type         | name   | label |
                |        | csv-external | mydata |       |
                |        | begin group  | g1     |       |
                |        | xml-external | mydata |       |
                |        | end group    | g1     |       |
                """,
                model__contains=[],
            )
        self.assertIn("Instance names must be unique", repr(ctx.exception))
        self.assertIn("The name 'mydata' was found 2 time(s)", repr(ctx.exception))

    def test_can__use_unique_external_xml_across_groups(self):
        """Unique external instances anywhere is OK."""
        self.assertPyxformXform(
            md="""
            | survey |              |         |                |
            |        | type         | name    | label          |
            |        | xml-external | mydata  |                |
            |        | begin group  | g1      |                |
            |        | xml-external | mydata1 |                |
            |        | note         | note1   | It's note-able |
            |        | end group    | g1      |                |
            |        | begin group  | g2      |                |
            |        | note         | note2   | It's note-able |
            |        | xml-external | mydata2 |                |
            |        | end group    | g2      |                |
            |        | begin group  | g3      |                |
            |        | note         | note3   | It's note-able |
            |        | xml-external | mydata3 |                |
            |        | end group    | g3      |                |
            """,
            model__contains=[
                '<instance id="mydata" src="jr://file/mydata.xml"/>',
                '<instance id="mydata1" src="jr://file/mydata1.xml"/>',
                '<instance id="mydata2" src="jr://file/mydata2.xml"/>',
                '<instance id="mydata3" src="jr://file/mydata3.xml"/>',
            ],
        )

    def test_cannot__use_same_external_xml_id_with_mixed_types(self):
        """Duplicate instances with other sources present raises an error."""
        with self.assertRaises(PyxformTestError) as ctx:
            self.assertPyxformXform(
                md="""
                | survey |                                      |      |       |                                             |
                |        | type                                 | name | label | calculation                                 |
                |        | begin group                          | g1   |       |                                             |
                |        | xml-external                         | city |       |                                             |
                |        | end group                            | g1   |       |                                             |
                |        | xml-external                         | city |       |                                             |
                |        | begin group                          | g2   |       |                                             |
                |        | select_one_from_file cities.csv      | city | City  |                                             |
                |        | end group                            | g2   |       |                                             |
                |        | begin group                          | g3   |       |                                             |
                |        | select_multiple_from_file cities.csv | city | City  |                                             |
                |        | end group                            | g3   |       |                                             |
                |        | begin group                          | g4   |       |                                             |
                |        | calculate                            | city | City  | pulldata('fruits', 'name', 'name', 'mango') |
                |        | end group                            | g4   |       |                                             |
                """,  # noqa
                model__contains=[],
            )
        self.assertIn("The name 'city' was found 2 time(s)", repr(ctx.exception))

    def test_can__use_same_external_csv_id_with_mixed_types(self):
        """Multiple fields that require the same external instance result in a single instance declaration."""
        self.assertPyxformXform(
            md="""
            | survey |                                      |      |       |                                             |
            |        | type                                 | name | label | calculation                                 |
            |        | begin group                          | g1   |       |                                             |
            |        | text                                 | foo  | Foo   |                                             |
            |        | csv-external                         | city |       |                                             |
            |        | end group                            | g1   |       |                                             |
            |        | begin group                          | g2   |       |                                             |
            |        | select_one_from_file cities.csv      | city | City  |                                             |
            |        | end group                            | g2   |       |                                             |
            |        | begin group                          | g3   |       |                                             |
            |        | select_multiple_from_file cities.csv | city | City  |                                             |
            |        | end group                            | g3   |       |                                             |
            |        | begin group                          | g4   |       |                                             |
            |        | text                                 | foo  | Foo   |                                             |
            |        | calculate                            | city | City  | pulldata('fruits', 'name', 'name', 'mango') |
            |        | end group                            | g4   |       |                                             |
            """,  # noqa
            model__contains=['<instance id="city" src="jr://file-csv/city.csv"/>'],
        )

    def test_can__use_all_types_together_with_unique_ids(self):
        """Unique instances with other sources present are OK."""
        self.assertPyxformXform(
            md="""
            | survey  |                                      |       |       |                                             |               |
            |         | type                                 | name  | label | calculation                                 | choice_filter |
            |         | begin group                          | g1    |       |                                             |               |
            |         | xml-external                         | city1 |       |                                             |               |
            |         | note                                 | note1 | Note  |                                             |               |
            |         | end group                            | g1    |       |                                             |               |
            |         | begin group                          | g2    |       |                                             |               |
            |         | select_one_from_file cities.csv      | city2 | City2 |                                             |               |
            |         | end group                            | g2    |       |                                             |               |
            |         | begin group                          | g3    |       |                                             |               |
            |         | select_multiple_from_file cities.csv | city3 | City3 |                                             |               |
            |         | end group                            | g3    |       |                                             |               |
            |         | begin group                          | g4    |       |                                             |               |
            |         | calculate                            | city4 | City4 | pulldata('fruits', 'name', 'name', 'mango') |               |
            |         | note                                 | note4 | Note  |                                             |               |
            |         | end group                            | g4    |       |                                             |               |
            |         | select_one states                    | test  | Test  |                                             | true()        |
            | choices |                                      |       |       |                                             |               |
            |         | list_name                            | name  | label |                                             |               |
            |         | states                               | 1     | Pass  |                                             |               |
            |         | states                               | 2     | Fail  |                                             |               |
            """,  # noqa
            model__contains=[
                '<instance id="city1" src="jr://file/city1.xml"/>',
                '<instance id="cities" src="jr://file-csv/cities.csv"/>',
                '<instance id="fruits" src="jr://file-csv/fruits.csv"/>',
                """
      <instance id="states">
        <root>
          <item>
            <label>Pass</label>
            <name>1</name>
          </item>
          <item>
            <label>Fail</label>
            <name>2</name>
          </item>
        </root>
      </instance>
""",
            ],  # noqa
        )

    def test_cannot__use_different_src_same_id__select_then_internal(self):
        """Duplicate instance from internal choices raises an error: #88."""
        md = """
            | survey  |                                 |       |       |                                    |
            |         | type                            | name  | label | choice_filter                      |
            |         | select_one_from_file states.csv | state | State |                                    |
            |         | select_one states               | test  | Test  | state=/select_from_file_test/state |
            | choices |                                 |       |       |                                    |
            |         | list_name                       | name  | label |                                    |
            |         | states                          | 1     | Pass  |                                    |
            |         | states                          | 2     | Fail  |                                    |
            """  # noqa
        with self.assertRaises(PyXFormError) as ctx:
            survey = self.md_to_pyxform_survey(md_raw=md)
            survey._to_pretty_xml()
        self.assertIn(
            "Instance name: 'states', "
            "Existing type: 'file', Existing URI: 'jr://file-csv/states.csv', "
            "Duplicate type: 'choice', Duplicate URI: 'None', "
            "Duplicate context: 'survey'.",
            repr(ctx.exception),
        )

    def test_cannot__use_different_src_same_id__external_then_pulldata(self):
        """
        Duplicate instance from pulldata after xml-external raises an error.
        """
        md = """
            | survey |              |        |                  |                                             |
            |        | type         | name   | label            | calculation                                 |
            |        | begin group  | g1     |                  |                                             |
            |        | xml-external | fruits |                  |                                             |
            |        | calculate    | f_csv  | City             | pulldata('fruits', 'name', 'name', 'mango') |
            |        | note         | note   | Fruity! ${f_csv} |                                             |
            |        | end group    | g1     |                  |                                             |
            """  # noqa
        with self.assertRaises(PyXFormError) as ctx:
            survey = self.md_to_pyxform_survey(md_raw=md)
            survey._to_pretty_xml()
        self.assertIn(
            "Instance name: 'fruits', "
            "Existing type: 'external', Existing URI: 'jr://file/fruits.xml', "
            "Duplicate type: 'pulldata', Duplicate URI: 'jr://file-csv/fruits.csv', "  # noqa
            "Duplicate context: '[type: group, name: g1]'.",
            repr(ctx.exception),
        )

    def test_cannot__use_different_src_same_id__pulldata_then_external(self):
        """
        Duplicate instance from xml-external after pulldata raises an error.
        """
        md = """
            | survey |              |        |                  |                                             |
            |        | type         | name   | label            | calculation                                 |
            |        | begin group  | g1     |                  |                                             |
            |        | calculate    | f_csv  | City             | pulldata('fruits', 'name', 'name', 'mango') |
            |        | xml-external | fruits |                  |                                             |
            |        | note         | note   | Fruity! ${f_csv} |                                             |
            |        | end group    | g1     |                  |                                             |
            """  # noqa
        with self.assertRaises(PyXFormError) as ctx:
            survey = self.md_to_pyxform_survey(md_raw=md)
            survey._to_pretty_xml()
        self.assertIn(
            "Instance name: 'fruits', "
            "Existing type: 'pulldata', Existing URI: 'jr://file-csv/fruits.csv', "  # noqa
            "Duplicate type: 'external', Duplicate URI: 'jr://file/fruits.xml', "  # noqa
            "Duplicate context: '[type: group, name: g1]'.",
            repr(ctx.exception),
        )

    def test_can__reuse_csv__selects_then_pulldata(self):
        """Re-using the same csv external data source id and URI is OK."""
        md = """
            | survey |                                              |        |                                    |                                                   |
            |        | type                                         | name   | label                              | calculation                                       |
            |        | select_multiple_from_file pain_locations.csv | plocs  | Locations of pain this week.       |                                                   |
            |        | select_one_from_file pain_locations.csv      | pweek  | Location of worst pain this week.  |                                                   |
            |        | select_one_from_file pain_locations.csv      | pmonth | Location of worst pain this month. |                                                   |
            |        | select_one_from_file pain_locations.csv      | pyear  | Location of worst pain this year.  |                                                   |
            |        | calculate                                    | f_csv  | pd                                 | pulldata('pain_locations', 'name', 'name', 'arm') |
            |        | note                                         | note   | Arm ${f_csv}                       |                                                   |
            """  # noqa
        expected = """
      <instance id="pain_locations" src="jr://file-csv/pain_locations.csv"/>
"""  # noqa
        self.assertPyxformXform(md=md, model__contains=[expected])
        survey = self.md_to_pyxform_survey(md_raw=md)
        xml = survey._to_pretty_xml()
        self.assertEqual(1, xml.count(expected))

    def test_can__reuse_csv__pulldata_then_selects(self):
        """Re-using the same csv external data source id and URI is OK."""
        md = """
            | survey |                                              |        |                                    |                                                   |
            |        | type                                         | name   | label                              | calculation                                       |
            |        | calculate                                    | f_csv  | pd                                 | pulldata('pain_locations', 'name', 'name', 'arm') |
            |        | note                                         | note   | Arm ${f_csv}                       |                                                   |
            |        | select_multiple_from_file pain_locations.csv | plocs  | Locations of pain this week.       |                                                   |
            |        | select_one_from_file pain_locations.csv      | pweek  | Location of worst pain this week.  |                                                   |
            |        | select_one_from_file pain_locations.csv      | pmonth | Location of worst pain this month. |                                                   |
            |        | select_one_from_file pain_locations.csv      | pyear  | Location of worst pain this year.  |                                                   |
            """  # noqa
        expected = """<instance id="pain_locations" src="jr://file-csv/pain_locations.csv"/>"""  # noqa
        self.assertPyxformXform(md=md, model__contains=[expected])

    def test_can__reuse_xml__selects_then_external(self):
        """Re-using the same xml external data source id and URI is OK."""
        md = """
            | survey |                                              |                |                                    |
            |        | type                                         | name           | label                              |
            |        | select_multiple_from_file pain_locations.xml | plocs          | Locations of pain this week.       |
            |        | select_one_from_file pain_locations.xml      | pweek          | Location of worst pain this week.  |
            |        | select_one_from_file pain_locations.xml      | pmonth         | Location of worst pain this month. |
            |        | select_one_from_file pain_locations.xml      | pyear          | Location of worst pain this year.  |
            |        | xml-external                                 | pain_locations |                                    |
            """  # noqa
        expected = """
      <instance id="pain_locations" src="jr://file/pain_locations.xml"/>
"""  # noqa
        survey = self.md_to_pyxform_survey(md_raw=md)
        xml = survey._to_pretty_xml()
        self.assertEqual(1, xml.count(expected))

    def test_can__reuse_xml__external_then_selects(self):
        """Re-using the same xml external data source id and URI is OK."""
        md = """
            | survey |                                              |                |                                    |
            |        | type                                         | name           | label                              |
            |        | xml-external                                 | pain_locations |                                    |
            |        | select_multiple_from_file pain_locations.xml | plocs          | Locations of pain this week.       |
            |        | select_one_from_file pain_locations.xml      | pweek          | Location of worst pain this week.  |
            |        | select_one_from_file pain_locations.xml      | pmonth         | Location of worst pain this month. |
            |        | select_one_from_file pain_locations.xml      | pyear          | Location of worst pain this year.  |
            """  # noqa
        expected = """<instance id="pain_locations" src="jr://file/pain_locations.xml"/>"""  # noqa
        self.assertPyxformXform(md=md, model__contains=[expected])
        survey = self.md_to_pyxform_survey(md_raw=md)
        xml = survey._to_pretty_xml()
        self.assertEqual(1, xml.count(expected))

    def test_external_instance_pulldata_constraint(self):
        """
        Checks if instance node for pulldata function is added
        when pulldata occurs in column with constraint title
        """
        md = """
        | survey |        |         |                |                                                         |
        |        | type   | name    | label          | constraint                                              |
        |        | text   | Part_ID | Participant ID | pulldata('ID', 'ParticipantID', 'ParticipantIDValue',.) |
        """
        node = """<instance id="ID" src="jr://file-csv/ID.csv"/>"""
        self.assertPyxformXform(md=md, xml__contains=[node])

    def test_external_instance_pulldata_constraint_in_expression(self):
        self.assertPyxformXform(
            md="""
            | survey |        |         |                |                                                             |
            |        | type   | name    | label          | constraint                                                  |
            |        | text   | Part_ID | Participant ID | . > pulldata('ID', 'ParticipantID', 'ParticipantIDValue',.) |
            """,
            xml__contains=["""<instance id="ID" src="jr://file-csv/ID.csv"/>"""],
        )

    def test_pulldata_calculate_multi_line_expression__one_call(self):
        """Should find the pulldata instance name in a multi-line expression."""
        qd = """\
          if(${a}, 
          pulldata('my_data_b', 'my_ref', 'metainstanceID', ${b}),
          (count-selected(${c})))
        """
        self.assertPyxformXform(
            # Can't use md here because of multi-line calculate expression.
            ss_structure={
                "survey": [
                    {"type": "calculate", "name": "a", "label": "QA", "calculate": "1"},
                    {"type": "calculate", "name": "b", "label": "QB", "calculate": "1"},
                    {"type": "calculate", "name": "c", "label": "QC", "calculate": "1"},
                    {
                        "type": "calculate",
                        "name": "d",
                        "label": "QD",
                        "calculate": dedent(qd),
                    },
                ]
            },
            xml__contains=[
                """<instance id="my_data_b" src="jr://file-csv/my_data_b.csv"/>"""
            ],
            # Is Validate OK with the multi-line expression.
            run_odk_validate=True,
        )

    def test_pulldata_calculate_multi_line_expression__multiple_calls(self):
        """Should find the pulldata instance names in a multi-line expression."""
        qd = """\
          if(${a}, 
          pulldata('my_data_b', 'my_ref', 'metainstanceID', ${b}),
          pulldata('my_data_c', 'my_ref', 'metainstanceID', ${c}))
        """
        self.assertPyxformXform(
            # Can't use md here because of multi-line calculate expression.
            ss_structure={
                "survey": [
                    {"type": "calculate", "name": "a", "label": "QA", "calculate": "1"},
                    {"type": "calculate", "name": "b", "label": "QB", "calculate": "1"},
                    {"type": "calculate", "name": "c", "label": "QC", "calculate": "1"},
                    {
                        "type": "calculate",
                        "name": "d",
                        "label": "QD",
                        "calculate": dedent(qd),
                    },
                ]
            },
            xml__contains=[
                """<instance id="my_data_b" src="jr://file-csv/my_data_b.csv"/>""",
                """<instance id="my_data_c" src="jr://file-csv/my_data_c.csv"/>""",
            ],
            # Is Validate OK with the multi-line expression and multiple instances.
            run_odk_validate=True,
        )

    def test_pulldata_calculate_single_line_expression__multiple_calls(self):
        """Should find all pulldata instance names in an expression."""
        qd = (
            "if(${a}, pulldata('my_data_b', 'my_ref', 'metainstanceID', ${b}), "
            "pulldata('my_data_c', 'my_ref', 'metainstanceID', ${c}))"
        )
        self.assertPyxformXform(
            # Not using md here due to very long calculate expression.
            ss_structure={
                "survey": [
                    {"type": "calculate", "name": "a", "label": "QA", "calculate": "1"},
                    {"type": "calculate", "name": "b", "label": "QB", "calculate": "1"},
                    {"type": "calculate", "name": "c", "label": "QC", "calculate": "1"},
                    {"type": "calculate", "name": "d", "label": "QD", "calculate": qd},
                ]
            },
            xml__contains=[
                """<instance id="my_data_b" src="jr://file-csv/my_data_b.csv"/>""",
                """<instance id="my_data_c" src="jr://file-csv/my_data_c.csv"/>""",
            ],
            # Is Validate OK with the multiple instances.
            run_odk_validate=True,
        )

    def test_external_instance_pulldata_readonly(self):
        """
        Checks if instance node for pulldata function is added
        when pulldata occurs in column with readonly title
        """
        md = """
        | survey |        |         |                |                                                         |
        |        | type   | name    | label          | readonly                                                |
        |        | text   | Part_ID | Participant ID | pulldata('ID', 'ParticipantID', 'ParticipantIDValue',.) |
        """
        node = """<instance id="ID" src="jr://file-csv/ID.csv"/>"""

        self.assertPyxformXform(
            md=md,
            xml__contains=[node],
            run_odk_validate=False,  # Validate sees self references in readonly as circular but shouldn't
        )

    def test_external_instance_pulldata_required(self):
        """
        Checks if instance node for pulldata function is added
        when pulldata occurs in column with required title
        """
        md = """
        | survey |        |         |                |                                                         |
        |        | type   | name    | label          | required                                                |
        |        | text   | Part_ID | Participant ID | pulldata('ID', 'ParticipantID', 'ParticipantIDValue',.) |
        """
        node = """<instance id="ID" src="jr://file-csv/ID.csv"/>"""
        self.assertPyxformXform(
            md=md,
            xml__contains=[node],
            run_odk_validate=False,  # Validate sees self references in requireds as circular but shouldn't
        )

    def test_external_instance_pulldata_relevant(self):
        """
        Checks if instance node for pulldata function is added
        when pulldata occurs in column with relevant title
        """
        md = """
        | survey |        |         |                |                                                         |
        |        | type   | name    | label          | relevant                                                |
        |        | text   | Part_ID | Participant ID | pulldata('ID', 'ParticipantID', 'ParticipantIDValue',.) |
        """
        node = """<instance id="ID" src="jr://file-csv/ID.csv"/>"""
        self.assertPyxformXform(
            md=md,
            xml__contains=[node],
            run_odk_validate=False,  # Validate sees self references in relevants as circular but shouldn't
        )

    # This is not something that is recommended since pulldata and choice_filter both should use XPath predicates
    # behind the scenes but it should still be possible.
    def test_external_instance_pulldata_choice_filter(self):
        self.assertPyxformXform(
            md="""
            | survey |                |      |       |                                                                         |
            |        | type           | name | label | choice_filter                                                           |
            |        | select_one foo | foo  | Foo   | contains(name, pulldata('ID', 'ParticipantID', 'ParticipantIDValue',.)) |
            | choices|                |      |       |                                                                         |
            |        | list_name      | name | label |                                                                         |
            |        | foo            | a    | A     |                                                                         |
            """,
            xml__contains=["""<instance id="ID" src="jr://file-csv/ID.csv"/>"""],
        )

    def test_external_instance_pulldata_default(self):
        self.assertPyxformXform(
            md="""
            | survey |      |      |       |                                                         |
            |        | type | name | label | default                                                 |
            |        | text | foo  | Foo   | pulldata('ID', 'ParticipantID', 'ParticipantIDValue',.) |
            """,
            xml__contains=["""<instance id="ID" src="jr://file-csv/ID.csv"/>"""],
        )

    def test_external_instance_pulldata(self):
        """
        Checks that only one instance node for pulldata is created
        if pulldata function is present in at least one columns with
        the titles: constraint, relevant, required, readonly
        """
        md = """
        | survey |        |         |                |                                                         |                                                         |                                                         |
        |        | type   | name    | label          | relevant                                                | required                                                | constraint                                              |
        |        | text   | Part_ID | Participant ID | pulldata('ID', 'ParticipantID', 'ParticipantIDValue',.) | pulldata('ID', 'ParticipantID', 'ParticipantIDValue',.) | pulldata('ID', 'ParticipantID', 'ParticipantIDValue',.) |
        """
        node = """<instance id="ID" src="jr://file-csv/ID.csv"/>"""
        survey = self.md_to_pyxform_survey(md_raw=md)
        xml = survey._to_pretty_xml()
        self.assertEqual(1, xml.count(node))

    def test_external_instances_multiple_diff_pulldatas(self):
        """
        Checks that all instances for pulldata that needs creation
        are created
        The situation is if pulldata is present in 2 or more
        columns but pulling data from different csv files
        """
        md = """
        | survey |        |         |                |                                                        |                                                             |
        |        | type   | name    | label          | relevant                                               | required                                                    |
        |        | text   | Part_ID | Participant ID | pulldata('fruits', 'name', 'name_key', 'mango')        | pulldata('OtherID', 'ParticipantID', ParticipantIDValue',.) |
        """
        node1 = '<instance id="fruits" src="jr://file-csv/fruits.csv"/>'
        node2 = '<instance id="OtherID" src="jr://file-csv/OtherID.csv"/>'

        survey = self.md_to_pyxform_survey(md_raw=md)
        xml = survey._to_pretty_xml()
        self.assertEqual(1, xml.count(node1))
        self.assertEqual(1, xml.count(node2))

    def test_mixed_quotes_and_functions_in_pulldata(self):
        # re: https://github.com/XLSForm/pyxform/issues/398
        self.assertPyxformXform(
            name="pulldata",
            md="""
                | survey |             |            |       |                                                            |
                |        | type        | name       | label | calculation                                                |
                |        | text        | rcid       | ID    |                                                            |
                |        | calculate   | calculate1 |       | pulldata("instance1","first","rcid",concat("Foo",${rcid})) |
                |        | calculate   | calculate2 |       | pulldata('instance2',"last",'rcid',concat('RC',${rcid}))   |
                |        | calculate   | calculate3 |       | pulldata('instance3','envelope','rcid',"Bar")              |
                |        | calculate   | calculate4 |       | pulldata('instance4'          ,'envelope','rcid',"Bar")    |
                """,  # noqa
            xml__contains=[
                '<instance id="instance1" src="jr://file-csv/instance1.csv"/>',
                '<instance id="instance2" src="jr://file-csv/instance2.csv"/>',
                '<instance id="instance3" src="jr://file-csv/instance3.csv"/>',
                '<instance id="instance4" src="jr://file-csv/instance4.csv"/>',
            ],
        )



# -*- coding: utf-8 -*-
"""
Test external instance syntax

See also test_external_instances
"""
from pyxform.tests_v1.pyxform_test_case import PyxformTestCase


class ExternalCSVInstancesTest(PyxformTestCase):
    def test_external_csv_instances(self):
        # re: https://github.com/XLSForm/pyxform/issues/30
        self.assertPyxformXform(
            name="ecsv",
            md="""
            | survey |                                              |                |                |
            |        | type                                         | name           | label          |
            |        | select_one_from_file cities.csv              | city           | City           |
            |        | select_multiple_from_file neighbourhoods.csv | neighbourhoods | Neighbourhoods |
            """,  # noqa
            xml__contains=[
                '<instance id="cities" src="jr://file-csv/cities.csv"/>',
                '<select1 ref="/ecsv/city">',
                "<itemset nodeset=\"instance('cities')/root/item\">",
                '<instance id="neighbourhoods" src="jr://file-csv/neighbourhoods.csv"/>',  # noqa
                '<select ref="/ecsv/neighbourhoods">',
                "<itemset nodeset=\"instance('neighbourhoods')/root/item\">",
            ],
            run_odk_validate=True,
        )

    def test_external_csv_instances_w_choice_filter(self):
        # re: https://github.com/XLSForm/pyxform/issues/30
        self.assertPyxformXform(
            name="ecsv",
            md="""
            | survey |                                              |                |                |
            |        | type                                         | name           | label          | choice_filter |
            |        | select_one_from_file cities.csv              | city           | City           |               |
            |        | select_multiple_from_file neighbourhoods.csv | neighbourhoods | Neighbourhoods | city=${city}  |
            """,  # noqa
            xml__contains=[
                '<instance id="cities" src="jr://file-csv/cities.csv"/>',  # noqa
                '<select1 ref="/ecsv/city">',
                '<instance id="neighbourhoods" src="jr://file-csv/neighbourhoods.csv"/>',  # noqa
                '<select ref="/ecsv/neighbourhoods">',
                "<itemset nodeset=\"instance('neighbourhoods')/root/item[city= /ecsv/city ]\">",  # noqa
            ],
            run_odk_validate=True,
        )


class ExternalXMLInstancesTest(PyxformTestCase):
    def test_external_xml_instances(self):
        # re: https://github.com/XLSForm/pyxform/issues/30
        self.assertPyxformXform(
            name="exml",
            md="""
            | survey |                                              |                |                |
            |        | type                                         | name           | label          |
            |        | select_one_from_file cities.xml              | city           | City           |
            |        | select_multiple_from_file neighbourhoods.xml | neighbourhoods | Neighbourhoods |
            """,  # noqa
            xml__contains=[
                '<instance id="cities" src="jr://file/cities.xml"/>',  # noqa
                '<select1 ref="/exml/city">',
                "<itemset nodeset=\"instance('cities')/root/item\">",
                '<instance id="neighbourhoods" src="jr://file/neighbourhoods.xml"/>',  # noqa
                '<select ref="/exml/neighbourhoods">',
                "<itemset nodeset=\"instance('neighbourhoods')/root/item\">",
            ],
            run_odk_validate=True,
        )


class InvalidExternalFileInstancesTest(PyxformTestCase):
    def test_external_other_extension_instances(self):
        # re: https://github.com/XLSForm/pyxform/issues/30
        self.assertPyxformXform(
            name="epdf",
            md="""
            | survey |                                              |                |                |
            |        | type                                         | name           | label          |
            |        | select_one_from_file cities.pdf              | city           | City           |
            |        | select_multiple_from_file neighbourhoods.pdf | neighbourhoods | Neighbourhoods |
            """,  # noqa
            errored=True,
            error_contains=["should be a choices sheet in this xlsform"],
        )

    def test_external_choices_sheet_included_instances(self):
        # re: https://github.com/XLSForm/pyxform/issues/30
        self.assertPyxformXform(
            name="epdf",
            md="""
            | survey |                                              |                |                |
            |        | type                                         | name           | label          |
            |        | select_one_from_file cities.pdf              | city           | City           |
            |        | select_multiple_from_file neighbourhoods.pdf | neighbourhoods | Neighbourhoods |

            | choices |
            |         | list name | name  | label |
            |         | fruits    | apple | Apple |
            """,  # noqa
            errored=True,
            error__contains=["List name not in choices sheet: cities.pdf"],
        )


class ExternalCSVInstancesBugsTest(PyxformTestCase):
    def test_non_existent_itext_reference(self):
        # re: https://github.com/XLSForm/pyxform/issues/80
        self.assertPyxformXform(
            name="ecsv",
            md="""
            | survey |                                              |                |                |
            |        | type                                         | name           | label          |
            |        | select_one_from_file cities.csv              | city           | City           |
            |        | select_multiple_from_file neighbourhoods.csv | neighbourhoods | Neighbourhoods |
            """,  # noqa
            xml__contains=[
                """<itemset nodeset="instance('cities')/root/item">
        <value ref="name"/>
        <label ref="label"/>
      </itemset>"""
            ],
        )

    def test_external_choices_sheet_values_required(self):
        md = """
            | survey           |                               |          |           |                                     |
            |                  | type                          | name     | label     | choice_filter                       |
            |                  | text                          | S1       | s1        |                                     |
            |                  | select_one_external counties  | county   | county    | S1=${S1}                            |
            |                  | select_one_external cities    | city     | city      | S1=${S1} and county=${county}       |
            | choices          |                               |          |           |
            |                  | list name                     | name     | label     |
            |                  | list                          | option a | a         |
            | external_choices |              |                |
            |                  | list_name    | name           |
            |                  | counties     | Kajiado        |
            |                  | counties     | Nakuru         |
            |                  | cities       | Kisumu         |
            |                  | cities       | Mombasa        |
            """
        expected = """
    <input query="instance('cities')/root/item[S1= /pyxform_autotestname/S1  and county= /pyxform_autotestname/county ]" ref="/pyxform_autotestname/city">
"""

        self.assertPyxformXform(
            md=md, debug=False, xml__contains=[expected], run_odk_validate=True
        )

    def test_list_name_not_in_external_choices_sheet_raises_error(self):
        self.assertPyxformXform(
            md="""
                | survey           |                              |                |           |                                     |
                |                  | type                         | name           | label     | choice_filter                       |
                |                  | select_one list              | S1             | s1        |                                     |
                |                  | select_one_external counties | county         | County    | S1=${S1}                            |
                |                  | select_one_external cities   | city           | City      | S1=${S1} and county=${county}       |
                | choices          |                              |                |           |
                |                  | list name                    | name           | label     |
                |                  | list                         | option a       | a         |
                |                  | list                         | option b       | b         |
                |                  | list                         | option c       | c         |
                | external_choices |                              |                |
                |                  | list_name                    | name           |
                |                  | counties                     | Kajiado        |
                |                  | counties                     | Nakuru         |
                """,  # noqa
            errored=True,
            error__contains=["List name not in external choices sheet: cities"],
        )



# -*- coding: utf-8 -*-
"""
This file contains constants that correspond with the property names in the
json survey format. (@see json_form_schema.json) These names are to be shared
between X2json and json2Y programs. By putting them in a shared file,
the literal names can be easily changed, typos can be avoided, and references
are easier to find.
"""
# TODO: Replace matching strings in the json2xforms code (builder.py,
# survey.py, survey_element.py, question.py) with these constants

TYPE = "type"
TITLE = "title"
NAME = "name"
ID_STRING = "id_string"
SMS_KEYWORD = "sms_keyword"
SMS_FIELD = "sms_field"
SMS_OPTION = "sms_option"
SMS_SEPARATOR = "sms_separator"
SMS_ALLOW_MEDIA = "sms_allow_media"
SMS_DATE_FORMAT = "sms_date_format"
SMS_DATETIME_FORMAT = "sms_datetime_format"
SMS_RESPONSE = "sms_response"

# compact representation (https://opendatakit.github.io/xforms-spec/#compact-record-representation-(for-sms))
COMPACT_PREFIX = "prefix"
COMPACT_DELIMITER = "delimiter"
COMPACT_TAG = "compact_tag"

VERSION = "version"
PUBLIC_KEY = "public_key"
SUBMISSION_URL = "submission_url"
AUTO_SEND = "auto_send"
AUTO_DELETE = "auto_delete"
DEFAULT_LANGUAGE = "default_language"
LABEL = "label"
HINT = "hint"
STYLE = "style"
ATTRIBUTE = "attribute"
ALLOW_CHOICE_DUPLICATES = "allow_choice_duplicates"

BIND = "bind"  # TODO: What should I do with the nested types? (readonly and relevant) # noqa
MEDIA = "media"
CONTROL = "control"
APPEARANCE = "appearance"

LOOP = "loop"
COLUMNS = "columns"

REPEAT = "repeat"
GROUP = "group"
CHILDREN = "children"

SELECT_ONE = "select one"
SELECT_ALL_THAT_APPLY = "select all that apply"
RANK = "rank"
CHOICES = "choices"

# XLS Specific constants
LIST_NAME = "list name"
CASCADING_SELECT = "cascading_select"
TABLE_LIST = "table-list"  # hyphenated because it goes in appearance, and convention for appearance column is dashes # noqa
FIELD_LIST = "field-list"

SURVEY = "survey"
SETTINGS = "settings"
EXTERNAL_CHOICES = "external_choices"

OSM = "osm"
OSM_TYPE = "binary"

NAMESPACES = "namespaces"

# The following are the possible sheet names:
SUPPORTED_SHEET_NAMES = [
    SURVEY,
    CHOICES,
    SETTINGS,
    EXTERNAL_CHOICES,
    OSM,
]
SUPPORTED_FILE_EXTENSIONS = [".xls", ".xlsx", ".xlsm"]

LOCATION_PRIORITY = "location-priority"
LOCATION_MIN_INTERVAL = "location-min-interval"
LOCATION_MAX_AGE = "location-max-age"
TRACK_CHANGES = "track-changes"
IDENTIFY_USER = "identify-user"
TRACK_CHANGES_REASONS = "track-changes-reasons"

# supported bind keywords for which external instances will be created for pulldata function
EXTERNAL_INSTANCES = ["calculate", "constraint", "readonly", "required", "relevant"]

# The ODK XForms version that generated forms comply to
CURRENT_XFORMS_VERSION = "1.0.0"

DEPRECATED_DEVICE_ID_METADATA_FIELDS = ["subscriberid", "simserial"]

AUDIO_QUALITY_VOICE_ONLY = "voice-only"
AUDIO_QUALITY_LOW = "low"
AUDIO_QUALITY_NORMAL = "normal"
AUDIO_QUALITY_EXTERNAL = "external"



# -*- coding: utf-8 -*-
"""
PyxformTestCase base class using markdown to define the XLSForm.
"""
from __future__ import print_function, unicode_literals

import codecs
import logging
import os
import re
import tempfile
import xml.etree.ElementTree as ETree
from unittest import TestCase

from pyxform.builder import create_survey_element_from_dict
from pyxform.errors import PyXFormError
from pyxform.tests_v1.test_utils.md_table import md_table_to_ss_structure
from pyxform.utils import NSMAP, unicode
from pyxform.validators.odk_validate import ODKValidateError, check_xform
from pyxform.xls2json import workbook_to_json

logger = logging.getLogger(__name__)
logger.addHandler(logging.StreamHandler())
logger.setLevel(logging.DEBUG)


class PyxformTestError(Exception):
    pass


class PyxformMarkdown(object):
    """Transform markdown formatted xlsform to a pyxform survey object"""

    def md_to_pyxform_survey(self, md_raw, kwargs=None, autoname=True, warnings=None):
        if kwargs is None:
            kwargs = {}
        if autoname:
            kwargs = self._autoname_inputs(kwargs)
        _md = []
        for line in md_raw.split("\n"):
            if re.match(r"^\s+#", line):
                # ignore lines which start with pound sign
                continue
            elif re.match(r"^(.*)(#[^|]+)$", line):
                # keep everything before the # outside of the last occurrence
                # of |
                _md.append(re.match(r"^(.*)(#[^|]+)$", line).groups()[0].strip())
            else:
                _md.append(line.strip())
        md = "\n".join(_md)

        if kwargs.get("debug"):
            logger.debug(md)

        def list_to_dicts(arr):
            headers = arr[0]

            def _row_to_dict(row):
                out_dict = {}
                for i in range(0, len(row)):
                    col = row[i]
                    if col not in [None, ""]:
                        out_dict[headers[i]] = col
                return out_dict

            return [_row_to_dict(r) for r in arr[1:]]

        sheets = {}
        for sheet, contents in md_table_to_ss_structure(md):
            sheets[sheet] = list_to_dicts(contents)

        return self._ss_structure_to_pyxform_survey(sheets, kwargs, warnings=warnings)

    @staticmethod
    def _ss_structure_to_pyxform_survey(ss_structure, kwargs, warnings=None):
        # using existing methods from the builder
        imported_survey_json = workbook_to_json(ss_structure, warnings=warnings)
        # ideally, when all these tests are working, this would be
        # refactored as well
        survey = create_survey_element_from_dict(imported_survey_json)
        survey.name = kwargs.get("name", "data")
        survey.title = kwargs.get("title")
        survey.id_string = kwargs.get("id_string")

        return survey

    @staticmethod
    def _run_odk_validate(xml):
        # On Windows, NamedTemporaryFile must be opened exclusively.
        # So it must be explicitly created, opened, closed, and removed
        tmp = tempfile.NamedTemporaryFile(suffix=".xml", delete=False)
        tmp.close()
        try:
            with codecs.open(tmp.name, mode="w", encoding="utf-8") as fp:
                fp.write(xml)
                fp.close()
            check_xform(tmp.name)
        finally:
            # Clean up the temporary file
            os.remove(tmp.name)
            assert not os.path.isfile(tmp.name)

    @staticmethod
    def _autoname_inputs(kwargs):
        """
        title and name are necessary for surveys, but not always convenient to
        include in test cases, so this will pull a default value
        from the stack trace.
        """
        test_name_root = "pyxform"
        if "name" not in kwargs.keys():
            kwargs["name"] = test_name_root + "_autotestname"
        if "title" not in kwargs.keys():
            kwargs["title"] = test_name_root + "_autotesttitle"
        if "id_string" not in kwargs.keys():
            kwargs["id_string"] = test_name_root + "_autotest_id_string"

        return kwargs


class PyxformTestCase(PyxformMarkdown, TestCase):
    def assertPyxformXform(self, **kwargs):
        """
        PyxformTestCase.assertPyxformXform() named arguments:
        -----------------------------------------------------

        one of these possible survey input types
          * md: (str) a markdown formatted xlsform (easy to read in code)
                [consider a plugin to help with formatting md tables,
                 e.g. https://github.com/vkocubinsky/SublimeTableEditor]
          * ss_structure: (dict) a python dictionary with sheets and their
                contents. best used in cases where testing whitespace and
                cells' type is important
          * survey: (pyxform.survey.Survey) easy for reuse within a test
          # Note: XLS is not implemented at this time. You can use builder to
          create a pyxform Survey object

        one or many of these string "matchers":
          * xml__contains: an array of strings which exist in the
                resulting xml. [xml|model|instance|itext]_excludes are also supported.
          * error__contains: a list of strings which should exist in the error
          * error__not_contains: a list of strings which should not exist in the error
          * odk_validate_error__contains: list of strings; run_odk_validate must be set
          * warning__contains: a list of strings which should exist in the warnings
          * warning__not_contains: a list of strings which should not exist in the warnings
          * warnings_count: the number of expected warning messages
          * xml__excludes: an array of strings which should not exist in the resulting
               xml. [xml|model|instance|itext]_excludes are also supported.

        optional other parameters passed to pyxform:
          * errored: (bool) if the xlsform is not supposed to compile,
                this must be True
          * name: (str) a valid xml tag to be used as the form name
          * id_string: (str)
          * title: (str)
          * run_odk_validate: (bool) when True, runs ODK Validate process
                Default value = False because it slows down tests
          * warnings: (list) a list to use for storing warnings for inspection.
        """
        debug = kwargs.get("debug", False)
        expecting_invalid_survey = kwargs.get("errored", False)
        errors = []
        warnings = kwargs.get("warnings", [])
        xml_nodes = {}

        run_odk_validate = kwargs.get("run_odk_validate", False)
        odk_validate_error__contains = kwargs.get("odk_validate_error__contains", [])

        try:
            if "md" in kwargs.keys():
                kwargs = self._autoname_inputs(kwargs)
                survey = self.md_to_pyxform_survey(
                    kwargs.get("md"), kwargs, warnings=warnings
                )
            elif "ss_structure" in kwargs.keys():
                kwargs = self._autoname_inputs(kwargs)
                survey = self._ss_structure_to_pyxform_survey(
                    kwargs.get("ss_structure"), kwargs, warnings=warnings,
                )
            else:
                survey = kwargs.get("survey")

            xml = survey._to_pretty_xml()
            root = ETree.fromstring(xml.encode("utf-8"))

            # Ensure all namespaces are present, even if unused
            final_nsmap = NSMAP.copy()
            final_nsmap.update(survey.get_nsmap())
            root.attrib.update(final_nsmap)

            xml_nodes["xml"] = root

            def _pull_xml_node_from_root(element_selector):
                _r = root.findall(
                    ".//n:%s" % element_selector,
                    namespaces={"n": "http://www.w3.org/2002/xforms"},
                )
                if _r:
                    return _r[0]

                return False

            for _n in ["model", "instance", "itext"]:
                xml_nodes[_n] = _pull_xml_node_from_root(_n)
            if debug:
                logger.debug(xml)
            if run_odk_validate:
                self._run_odk_validate(xml=xml)
                if odk_validate_error__contains:
                    raise PyxformTestError("ODKValidateError was not raised")

        except PyXFormError as e:
            survey = False
            errors = [unicode(e)]
            if debug:
                logger.debug("<xml unavailable />")
                logger.debug("ERROR: '%s'", errors[0])
        except ODKValidateError as e:
            if not odk_validate_error__contains:
                raise PyxformTestError(
                    "ODK Validate error was thrown but "
                    + "'odk_validate_error__contains'"
                    + " was empty:"
                    + unicode(e)
                )
            for v_err in odk_validate_error__contains:
                self.assertContains(
                    e.args[0], v_err, msg_prefix="odk_validate_error__contains"
                )
        else:
            survey = True

        if survey:

            def _check(keyword, verb):
                verb_str = "%s__%s" % (keyword, verb)

                bad_kwarg = "%s_%s" % (code, verb)
                if bad_kwarg in kwargs:
                    good_kwarg = "%s__%s" % (code, verb)
                    raise SyntaxError(
                        (
                            "'%s' is not a valid parameter. "
                            "Use double underscores: '%s'"
                        )
                        % (bad_kwarg, good_kwarg)
                    )

                def check_content(content):
                    text_arr = kwargs[verb_str]
                    for i in text_arr:
                        if verb == "contains":
                            self.assertContains(content, i, msg_prefix=keyword)
                        else:
                            self.assertNotContains(content, i, msg_prefix=keyword)

                return verb_str, check_content

            if "body_contains" in kwargs or "body__contains" in kwargs:
                raise SyntaxError(
                    "Invalid parameter: 'body__contains'." "Use 'xml__contains' instead"
                )

            # guarantee that strings contain alphanumerically sorted attributes across Python versions
            reorder_attributes(root)

            for code in ["xml", "instance", "model", "itext"]:
                for verb in ["contains", "excludes"]:
                    (code__str, checker) = _check(code, verb)
                    if kwargs.get(code__str):
                        checker(
                            ETree.tostring(xml_nodes[code], encoding="utf-8").decode(
                                "utf-8"
                            )
                        )

        if survey is False and expecting_invalid_survey is False:
            raise PyxformTestError(
                "Expected valid survey but compilation failed. "
                "Try correcting the error with 'debug=True', "
                "setting 'errored=True', "
                "and or optionally 'error__contains=[...]'"
                "\nError(s): " + "\n".join(errors)
            )
        elif survey and expecting_invalid_survey:
            raise PyxformTestError("Expected survey to be invalid.")

        search_test_kwargs = (
            "error__contains",
            "error__not_contains",
            "warnings__contains",
            "warnings__not_contains",
        )
        for k in search_test_kwargs:
            if k not in kwargs:
                continue
            if k.endswith("__contains"):
                assertion = self.assertContains
            elif k.endswith("__not_contains"):
                assertion = self.assertNotContains
            else:
                raise PyxformTestError("Unexpected search test kwarg: {}".format(k))
            if k.startswith("error"):
                joined = "\n".join(errors)
            elif k.startswith("warnings"):
                joined = "\n".join(warnings)
            else:
                raise PyxformTestError("Unexpected search test kwarg: {}".format(k))
            for text in kwargs[k]:
                assertion(joined, text, msg_prefix=k)
        if "warnings_count" in kwargs:
            c = kwargs.get("warnings_count")
            if not isinstance(c, int):
                PyxformTestError("warnings_count must be an integer.")
            self.assertEqual(c, len(warnings))

    @staticmethod
    def _assert_contains(content, text, msg_prefix):
        if msg_prefix:
            msg_prefix += ": "

        # Account for space in self-closing tags
        text_repr = repr(text)
        content = content.replace(" />", "/>")
        real_count = content.count(text)

        return text_repr, real_count, msg_prefix

    def assertContains(self, content, text, count=None, msg_prefix=""):
        """
        FROM: django source- testcases.py

        Asserts that ``text`` occurs ``count`` times in the content string.
        If ``count`` is None, the count doesn't matter - the assertion is
        true if the text occurs at least once in the content.
        """
        text_repr, real_count, msg_prefix = self._assert_contains(
            content, text, msg_prefix
        )

        if count is not None:
            self.assertEqual(
                real_count,
                count,
                msg_prefix + "Found %d instances of %s in content"
                " (expected %d)" % (real_count, text_repr, count),
            )
        else:
            self.assertTrue(
                real_count != 0,
                msg_prefix + "Couldn't find %s in content:\n" % text_repr + content,
            )

    def assertNotContains(self, content, text, msg_prefix=""):
        """
        Asserts that a content indicates that some content was retrieved
        successfully, (i.e., the HTTP status code was as expected), and that
        ``text`` doesn't occurs in the content of the content.
        """
        text_repr, real_count, msg_prefix = self._assert_contains(
            content, text, msg_prefix
        )

        self.assertEqual(
            real_count, 0, msg_prefix + "Response should not contain %s" % text_repr
        )


def reorder_attributes(root):
    """
    Forces alphabetical ordering of all XML attributes to match pre Python 3.8 behavior.
    In general, we should not rely on ordering, but changing all the tests is not
    realistic at this moment.

    See bottom of https://docs.python.org/3/library/xml.etree.elementtree.html#element-objects and
    https://github.com/python/cpython/commit/a3697db0102b9b6747fe36009e42f9b08f0c1ea8 for more information.

    NOTE: there's a similar ordering change made in utils.node. This one is also needed because in
    assertPyxformXform, the survey is converted to XML and then read back in using ETree.fromstring. This
    means that attribute ordering here is based on the attribute representation of xml.etree.ElementTree objects.
    In utils.node, it is based on xml.dom.minidom.Element objects. See https://github.com/XLSForm/pyxform/issues/414.
    """
    for el in root.iter():
        attrib = el.attrib
        if len(attrib) > 1:
            # Sort attributes. Attributes are represented as {namespace}name so attributes with explicit
            # namespaces will always sort after those without explicit namespaces.
            attribs = sorted(attrib.items())
            attrib.clear()
            attrib.update(attribs)



# -*- coding: utf-8 -*-
"""
Validate XForms using Enketo validator.
"""
import os
from typing import TYPE_CHECKING
from pyxform.validators.error_cleaner import ErrorCleaner
from pyxform.validators.util import (
    XFORM_SPEC_PATH,
    check_readable,
    decode_stream,
    run_popen_with_timeout,
)


if TYPE_CHECKING:
    from pyxform.validators.util import PopenResult


CURRENT_DIRECTORY = os.path.dirname(os.path.realpath(__file__))
ENKETO_VALIDATE_PATH = os.path.join(CURRENT_DIRECTORY, "bin", "validate")


class EnketoValidateError(Exception):
    """Common base class for Enketo validate exceptions."""

    pass


def install_exists():
    """
    Check if Enketo-validate is installed.
    """
    return os.path.exists(ENKETO_VALIDATE_PATH)


def _call_validator(path_to_xform, bin_file_path=ENKETO_VALIDATE_PATH) -> "PopenResult":
    return run_popen_with_timeout([bin_file_path, path_to_xform], 100)


def install_ok(bin_file_path=ENKETO_VALIDATE_PATH):
    """
    Check if Enketo-validate functions as expected.
    """
    check_readable(file_path=XFORM_SPEC_PATH)
    return_code, _, _, _ = _call_validator(
        path_to_xform=XFORM_SPEC_PATH, bin_file_path=bin_file_path
    )
    if return_code == 1:
        return False
    else:
        return True


def check_xform(path_to_xform):
    """
    Check the form with the Enketo validator.

    - return code 1: raise error with the stderr content.
    - return code 0: append warning with the stdout content (possibly none).

    :param path_to_xform: Path to the XForm to be validated.
    :return: warnings or List[str]
    """
    if not install_exists():
        raise EnvironmentError(
            "Enketo-validate dependency not found. "
            "Please use the updater tool to install the latest version."
        )

    returncode, timeout, stdout, stderr = _call_validator(path_to_xform=path_to_xform)
    warnings = []
    stderr = decode_stream(stderr)
    stdout = decode_stream(stdout)

    if timeout:
        return ["XForm took to long to completely validate."]
    else:
        if returncode > 0:  # Error invalid
            raise EnketoValidateError(
                "Enketo Validate Errors:\n" + ErrorCleaner.enketo_validate(stderr)
            )
        elif returncode == 0:
            if stdout:
                warnings.append("Enketo Validate Warnings:\n" + stdout)
            return warnings
        elif returncode < 0:
            return ["Bad return code from Enketo Validate."]
